"""Test ``seddy.app``."""

import json
from unittest import mock

from seddy import registration as seddy_registration
from seddy import _specs as seddy_decisions
import moto
import boto3
import pytest


class Workflow(seddy_decisions.Workflow):
    """Test workflow specification."""

    spec_type = "test"
    decisions_builder = None


seddy_decisions.WORKFLOW["test"] = Workflow


@moto.mock_swf
def test_list_workflows():
    """Test registered workflow listing."""
    # Setup environment
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(name="spam", workflowExecutionRetentionPeriodInDays="2")
    client.register_workflow_type(domain="spam", name="foo", version="1.0")
    client.register_workflow_type(domain="spam", name="foo", version="1.1")
    client.register_workflow_type(domain="spam", name="bar", version="0.42")
    client.deprecate_workflow_type(
        domain="spam", workflowType={"name": "foo", "version": "1.0"}
    )

    # Run function
    res = seddy_registration.list_workflows("spam", client)
    assert res == {("foo", "1.0"): False, ("foo", "1.1"): True, ("bar", "0.42"): True}


@moto.mock_swf
def test_register_workflow():
    """Test workflow registration."""
    # Setup environment
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(name="spam", workflowExecutionRetentionPeriodInDays="2")

    # Build input
    registration = seddy_decisions.Registration(
        task_timeout="NONE",
        execution_timeout=60,
        task_list="eggs",
        task_priority=2,
        child_policy=seddy_decisions.ChildPolicy.TERMINATE,
        lambda_role="arn:aws:iam::spam:role/eggs",
    )
    workflow = Workflow("foo", "0.42", "A workflow.", registration)

    # Run function
    seddy_registration.register_workflow(workflow, "spam", client)

    # Check registered workflow
    workflow_info = client.describe_workflow_type(
        domain="spam", workflowType={"name": "foo", "version": "0.42"}
    )
    assert workflow_info["typeInfo"]["status"] == "REGISTERED"
    assert workflow_info["typeInfo"]["description"] == "A workflow."
    cfg = workflow_info["configuration"]
    assert cfg["defaultTaskStartToCloseTimeout"] == "NONE"
    assert cfg["defaultExecutionStartToCloseTimeout"] == "60"
    assert cfg["defaultTaskList"] == {"name": "eggs"}
    assert cfg["defaultTaskPriority"] == "2"
    assert cfg["defaultChildPolicy"] == "TERMINATE"
    assert cfg["defaultLambdaRole"] == "arn:aws:iam::spam:role/eggs"


@moto.mock_swf
def test_deprecate_workflow():
    """Test workflow deprecation."""
    # Setup environment
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(name="spam", workflowExecutionRetentionPeriodInDays="2")
    client.register_workflow_type(domain="spam", name="foo", version="0.42")

    # Build input
    workflow = Workflow("foo", "0.42", "A workflow.")

    # Run function
    seddy_registration.deprecate_workflow(workflow, "spam", client)

    # Check registered workflow
    workflow_info = client.describe_workflow_type(
        domain="spam", workflowType={"name": "foo", "version": "0.42"}
    )
    assert workflow_info["typeInfo"]["status"] == "DEPRECATED"


@moto.mock_swf
def test_undeprecate_workflow():
    """Test workflow undeprecation."""
    # Setup environment
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(name="spam", workflowExecutionRetentionPeriodInDays="2")
    client.register_workflow_type(domain="spam", name="foo", version="0.42")
    client.deprecate_workflow_type(
        domain="spam", workflowType={"name": "foo", "version": "0.42"}
    )

    # Build input
    workflow = Workflow("foo", "0.42", "A workflow.")

    # Run function
    seddy_registration.undeprecate_workflow(workflow, "spam", client)

    # Check registered workflow
    workflow_info = client.describe_workflow_type(
        domain="spam", workflowType={"name": "foo", "version": "0.42"}
    )
    assert workflow_info["typeInfo"]["status"] == "REGISTERED"


@moto.mock_swf
def test_register_workflows():
    """Test workflows registration syncing."""
    # Setup environment
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(name="spam", workflowExecutionRetentionPeriodInDays="2")
    client.register_workflow_type(domain="spam", name="foo", version="1.0")
    client.register_workflow_type(domain="spam", name="foo", version="1.1")
    client.register_workflow_type(domain="spam", name="bar", version="0.42")
    client.register_workflow_type(domain="spam", name="bar", version="0.43")
    client.register_workflow_type(domain="spam", name="yay", version="0.17")
    client.register_workflow_type(domain="spam", name="yay", version="0.18")
    client.deprecate_workflow_type(
        domain="spam", workflowType={"name": "foo", "version": "1.0"}
    )
    client.deprecate_workflow_type(
        domain="spam", workflowType={"name": "bar", "version": "0.42"}
    )
    client.deprecate_workflow_type(
        domain="spam", workflowType={"name": "yay", "version": "0.17"}
    )

    client_patch = mock.patch.object(boto3, "client", lambda x, **_: {"swf": client}[x])

    # Build input
    workflows = [
        Workflow("foo", "1.0", registration=seddy_decisions.Registration(active=False)),
        Workflow("foo", "1.1", registration=seddy_decisions.Registration(active=False)),
        Workflow("foo", "1.2", registration=seddy_decisions.Registration(active=True)),
        Workflow("bar", "0.42", registration=seddy_decisions.Registration(active=True)),
        Workflow("bar", "0.43", registration=seddy_decisions.Registration(active=True)),
        Workflow("bar", "0.44", "", seddy_decisions.Registration(active=False)),
    ]

    # Build expectation
    exp_registered_workflow_types = [
        {"name": "yay", "version": "0.18"},
        {"name": "bar", "version": "0.42"},
        {"name": "bar", "version": "0.43"},
        {"name": "foo", "version": "1.2"},
    ]
    exp_deprecated_workflow_types = [
        {"name": "yay", "version": "0.17"},
        {"name": "foo", "version": "1.0"},
        {"name": "foo", "version": "1.1"},
    ]

    # Run function
    with client_patch:
        seddy_registration.register_workflows(workflows, "spam")

    # Check registered workflows
    resp = client.list_workflow_types(domain="spam", registrationStatus="REGISTERED")
    workflow_types = [w["workflowType"] for w in resp["typeInfos"]]
    workflow_types = sorted(workflow_types, key=lambda x: x["version"])
    assert workflow_types == exp_registered_workflow_types

    resp = client.list_workflow_types(domain="spam", registrationStatus="DEPRECATED")
    workflow_types = [w["workflowType"] for w in resp["typeInfos"]]
    workflow_types = sorted(workflow_types, key=lambda x: x["version"])
    assert workflow_types == exp_deprecated_workflow_types


def test_run_app(tmp_path):
    """Ensure workflow registration app is run correctly."""
    # Setup environment
    register_mock = mock.Mock()
    register_patch = mock.patch.object(
        seddy_registration, "register_workflows", register_mock
    )

    # Build input
    workflows_spec = {
        "version": "1.0",
        "workflows": [
            {"spec_type": "test", "name": "spam", "version": "1.0"},
            {
                "spec_type": "test",
                "name": "foo",
                "version": "0.42",
                "description": "The best workflow, bar none.",
                "registration": {
                    "task_timeout": "NONE",
                    "execution_timeout": 60,
                    "task_list": "eggs",
                    "task_priority": 2,
                    "child_policy": "TERMINATE",
                    "lambda_role": "arn:aws:iam::spam:role/eggs",
                },
            },
        ],
    }
    workflows_spec_json = tmp_path / "workflows.json"
    workflows_spec_json.write_text(json.dumps(workflows_spec, indent=4))

    # Build expectation
    registration1 = seddy_decisions.Registration(
        task_timeout="NONE",
        execution_timeout=60,
        task_list="eggs",
        task_priority=2,
        child_policy=seddy_decisions.ChildPolicy.TERMINATE,
        lambda_role="arn:aws:iam::spam:role/eggs",
    )
    workflows = [
        Workflow("spam", "1.0"),
        Workflow("foo", "0.42", "The best workflow, bar none.", registration1),
    ]

    # Run function
    with register_patch:
        seddy_registration.run_app(workflows_spec_json, "spam")

    # Check workflow registration configuration
    register_mock.assert_called_once_with(mock.ANY, "spam")

    res_workflows = register_mock.call_args_list[0][0][0]
    assert len(res_workflows) == len(workflows)
    for res_workflow, workflow in zip(res_workflows, workflows):
        assert isinstance(res_workflow, type(workflow))
        assert res_workflow.name == workflow.name
        assert res_workflow.version == workflow.version
        assert res_workflow.description == workflow.description
        if workflow.registration:
            registration = workflow.registration
            res_registration = res_workflow.registration
            assert res_registration.active == registration.active
            assert res_registration.task_timeout == registration.task_timeout
            assert res_registration.execution_timeout == registration.execution_timeout
            assert res_registration.task_list == registration.task_list
            assert res_registration.task_priority == registration.task_priority
            assert res_registration.child_policy == registration.child_policy
            assert res_registration.lambda_role == registration.lambda_role



"""SWF decisions making."""

import json
import string
import dataclasses
import typing as t
import logging as lg

from . import _base

logger = lg.getLogger(__name__)
_jsonpath_characters = string.digits + string.ascii_letters + "_"
_sentinel = object()
_attr_keys = {
    "ActivityTaskCancelRequested": "activityTaskCancelRequestedEventAttributes",
    "ActivityTaskCanceled": "activityTaskCanceledEventAttributes",
    "ActivityTaskCompleted": "activityTaskCompletedEventAttributes",
    "ActivityTaskFailed": "activityTaskFailedEventAttributes",
    "ActivityTaskScheduled": "activityTaskScheduledEventAttributes",
    "ActivityTaskStarted": "activityTaskStartedEventAttributes",
    "ActivityTaskTimedOut": "activityTaskTimedOutEventAttributes",
    "CancelTimerFailed": "cancelTimerFailedEventAttributes",
    "CancelWorkflowExecutionFailed": "cancelWorkflowExecutionFailedEventAttributes",
    "ChildWorkflowExecutionCanceled": "childWorkflowExecutionCanceledEventAttributes",
    "ChildWorkflowExecutionCompleted": "childWorkflowExecutionCompletedEventAttributes",
    "ChildWorkflowExecutionFailed": "childWorkflowExecutionFailedEventAttributes",
    "ChildWorkflowExecutionStarted": "childWorkflowExecutionStartedEventAttributes",
    "ChildWorkflowExecutionTerminated": "childWorkflowExecutionTerminatedEventAttributes",
    "ChildWorkflowExecutionTimedOut": "childWorkflowExecutionTimedOutEventAttributes",
    "CompleteWorkflowExecutionFailed": "completeWorkflowExecutionFailedEventAttributes",
    "ContinueAsNewWorkflowExecutionFailed": "continueAsNewWorkflowExecutionFailedEventAttributes",
    "DecisionTaskCompleted": "decisionTaskCompletedEventAttributes",
    "DecisionTaskScheduled": "decisionTaskScheduledEventAttributes",
    "DecisionTaskStarted": "decisionTaskStartedEventAttributes",
    "DecisionTaskTimedOut": "decisionTaskTimedOutEventAttributes",
    "ExternalWorkflowExecutionCancelRequested": "externalWorkflowExecutionCancelRequestedEventAttributes",
    "ExternalWorkflowExecutionSignaled": "externalWorkflowExecutionSignaledEventAttributes",
    "FailWorkflowExecutionFailed": "failWorkflowExecutionFailedEventAttributes",
    "LambdaFunctionCompleted": "lambdaFunctionCompletedEventAttributes",
    "LambdaFunctionFailed": "lambdaFunctionFailedEventAttributes",
    "LambdaFunctionScheduled": "lambdaFunctionScheduledEventAttributes",
    "LambdaFunctionStarted": "lambdaFunctionStartedEventAttributes",
    "LambdaFunctionTimedOut": "lambdaFunctionTimedOutEventAttributes",
    "MarkerRecorded": "markerRecordedEventAttributes",
    "RecordMarkerFailed": "recordMarkerFailedEventAttributes",
    "RequestCancelActivityTaskFailed": "requestCancelActivityTaskFailedEventAttributes",
    "RequestCancelExternalWorkflowExecutionFailed": "requestCancelExternalWorkflowExecutionFailedEventAttributes",
    "RequestCancelExternalWorkflowExecutionInitiated": "requestCancelExternalWorkflowExecutionInitiatedEventAttributes",
    "ScheduleActivityTaskFailed": "scheduleActivityTaskFailedEventAttributes",
    "ScheduleLambdaFunctionFailed": "scheduleLambdaFunctionFailedEventAttributes",
    "SignalExternalWorkflowExecutionFailed": "signalExternalWorkflowExecutionFailedEventAttributes",
    "SignalExternalWorkflowExecutionInitiated": "signalExternalWorkflowExecutionInitiatedEventAttributes",
    "StartChildWorkflowExecutionFailed": "startChildWorkflowExecutionFailedEventAttributes",
    "StartChildWorkflowExecutionInitiated": "startChildWorkflowExecutionInitiatedEventAttributes",
    "StartLambdaFunctionFailed": "startLambdaFunctionFailedEventAttributes",
    "StartTimerFailed": "startTimerFailedEventAttributes",
    "TimerCanceled": "timerCanceledEventAttributes",
    "TimerFired": "timerFiredEventAttributes",
    "TimerStarted": "timerStartedEventAttributes",
    "WorkflowExecutionCancelRequested": "workflowExecutionCancelRequestedEventAttributes",
    "WorkflowExecutionCanceled": "workflowExecutionCanceledEventAttributes",
    "WorkflowExecutionCompleted": "workflowExecutionCompletedEventAttributes",
    "WorkflowExecutionContinuedAsNew": "workflowExecutionContinuedAsNewEventAttributes",
    "WorkflowExecutionFailed": "workflowExecutionFailedEventAttributes",
    "WorkflowExecutionSignaled": "workflowExecutionSignaledEventAttributes",
    "WorkflowExecutionStarted": "workflowExecutionStartedEventAttributes",
    "WorkflowExecutionTerminated": "workflowExecutionTerminatedEventAttributes",
    "WorkflowExecutionTimedOut": "workflowExecutionTimedOutEventAttributes",
}
_error_events = {
    "ActivityTaskFailed",
    "ActivityTaskTimedOut",
    "CancelTimerFailed",
    "CancelWorkflowExecutionFailed",
    "CompleteWorkflowExecutionFailed",
    "DecisionTaskTimedOut",
    "FailWorkflowExecutionFailed",
    "RecordMarkerFailed",
    "RequestCancelActivityTaskFailed",
    "ScheduleActivityTaskFailed",
    "StartTimerFailed",
    "WorkflowExecutionCancelRequested",
}
_activity_events = {
    "ActivityTaskCompleted",
    "ActivityTaskFailed",
    "ActivityTaskTimedOut",
    "ActivityTaskScheduled",
    "ActivityTaskStarted",
}
_decision_failed_events = {
    "ScheduleActivityTaskFailed",
    "RequestCancelActivityTaskFailed",
    "StartTimerFailed",
    "CancelTimerFailed",
    "StartChildWorkflowExecutionFailed",
    "SignalExternalWorkflowExecutionFailed",
    "RequestCancelExternalWorkflowExecutionFailed",
    "CancelWorkflowExecutionFailed",
    "CompleteWorkflowExecutionFailed",
    "ContinueAsNewWorkflowExecutionFailed",
    "FailWorkflowExecutionFailed",
}


@dataclasses.dataclass
class TaskInput:
    @staticmethod
    def from_spec(spec: t.Dict[str, t.Any]) -> "TaskInput":
        for cls in [NoInput, Constant, WorkflowInput, DependencyResult, Object]:
            if cls.type == spec["type"]:
                break
        else:  # TODO: unit-test
            raise ValueError(spec["type"])
        return cls.from_spec(spec)


@dataclasses.dataclass
class NoInput(TaskInput):
    type: t.ClassVar = "none"

    @classmethod
    def from_spec(cls, spec) -> "NoInput":
        return cls()


@dataclasses.dataclass
class Constant(TaskInput):
    type: t.ClassVar = "constant"
    value: t.Any

    @classmethod
    def from_spec(cls, spec) -> "Constant":
        return cls(spec["value"])


@dataclasses.dataclass
class WorkflowInput(TaskInput):
    type: t.ClassVar = "workflow-input"
    path: str = "$"

    @classmethod
    def from_spec(cls, spec) -> "WorkflowInput":
        kwargs = {}
        if "path" in spec:
            kwargs["path"] = spec["path"]
        return cls(**kwargs)


@dataclasses.dataclass
class DependencyResult(TaskInput):
    type: t.ClassVar = "dependency-result"
    id: t.Any
    path: str = "$"

    @classmethod
    def from_spec(cls, spec) -> "DependencyResult":
        kwargs = {}
        if "path" in spec:
            kwargs["path"] = spec["path"]
        return cls(spec["id"], **kwargs)


@dataclasses.dataclass
class Object(TaskInput):
    type: t.ClassVar = "object"
    items: t.Dict[str, TaskInput]

    @classmethod
    def from_spec(cls, spec) -> "Object":
        items = {}
        for key, subspec in spec["items"].items():
            items[key] = cls.from_spec(subspec)
        return cls(items)


@dataclasses.dataclass
class Task:  # TODO: unit-test
    """DAG-type workflow activity task specification.

    Args:
        id: task ID, must be unique within a workflow execution and
            without ':', '/', '|', 'arn' or any control character
        name: activity type name
        version: activity type version
        heartbeat: task heartbeat time-out (seconds), or "NONE" for
            unlimited
        timeout: task time-out (seconds), or "None" for unlimited
        task_list: task-list to schedule task on
        priority: task priority
        dependencies: IDs of task’s dependencies
    """

    id: str
    name: str
    version: str
    input: t.Union[NoInput, Constant, WorkflowInput, DependencyResult, Object] = None
    heartbeat: t.Union[int, str] = None
    timeout: t.Union[int, str] = None
    task_list: str = None
    priority: int = None
    dependencies: t.List[str] = None
    _input_cls: t.ClassVar = TaskInput

    @property
    def type(self) -> t.Dict[str, str]:
        """Activity type."""
        return {"name": self.name, "version": self.version}

    @classmethod
    def from_spec(cls, spec: t.Dict[str, t.Any]) -> "Task":
        """Construct registration configuration from specification.

        Args:
            spec: workflow registration configuration specification
        """

        args = (spec["id"], spec["type"]["name"], spec["type"]["version"])
        kwargs = {}
        if "input" in spec:
            kwargs["input"] = cls._input_cls.from_spec(spec["input"])
        if "heartbeat" in spec:
            kwargs["heartbeat"] = spec["heartbeat"]
        if "timeout" in spec:
            kwargs["timeout"] = spec["timeout"]
        if "task_list" in spec:
            kwargs["task_list"] = spec["task_list"]
        if "priority" in spec:
            kwargs["priority"] = spec["priority"]
        if "dependencies" in spec:
            kwargs["dependencies"] = spec["dependencies"]
        return cls(*args, **kwargs)


def _get(item_id, items, id_key):
    """Get item from list with given ID."""
    return next(item for item in items if item[id_key] == item_id)


def _get_item_jsonpath(path: str, obj) -> t.Any:
    """Get a child item from an object.

    Args:
        path: path to child item, using basic single-valued JSONPath
            syntax
        obj: object to get child item from

    Returns:
        pointed-to child item

    Raises:
        ValueError: invalid path
    """

    if path[0] != "$":
        raise ValueError("invalid path (must start at root): %s" % path)

    indices = []
    chars = []
    state = None
    for char in path[1:]:
        if char in ".[":
            if state:
                if state != ".":
                    raise ValueError("invalid path (missing closing ']'): %s" % path)
                elif not chars:
                    raise ValueError("invalid path (empty key): %s" % path)
                indices.append("".join(chars))
            chars = []
            state = char
        elif char == "]":
            if state != "[":
                raise ValueError("invalid path (invalid key): %s" % path)
            elif not chars:
                raise ValueError("invalid path (empty key): %s" % path)
            index = int("".join(chars))
            indices.append(index)
            chars = []
            state = None
        elif char not in _jsonpath_characters:
            raise ValueError("invalid path (illegal characters): %s" % path)
        else:
            if not state:
                raise ValueError("invalid path (missing '.' or '['): %s" % path)
            chars.append(char)
    if state == "[":
        raise ValueError("invalid path (missing closing ']'): %s" % path)
    elif state == ".":
        indices.append("".join(chars))

    item = obj
    for index in indices:
        item = item[index]
    return item


def _build_activity_input(
    input_spec: TaskInput,
    workflow_input: t.Union[t.Dict[str, t.Any], None],
    activity_results: t.Dict[str, t.Any],
) -> t.Any:
    """Build activity input.

    Args:
        input_spec: activity task input specification
        workflow_input: workflow input
        activity_results: activities' results

    Returns:
        activity task input
    """

    input_spec = input_spec or NoInput()
    if isinstance(input_spec, NoInput):
        return _sentinel
    if isinstance(input_spec, Constant):
        return input_spec.value
    if isinstance(input_spec, WorkflowInput):
        path = input_spec.path
        return _get_item_jsonpath(path, workflow_input)
    if isinstance(input_spec, DependencyResult):
        path = input_spec.path
        dependency_result = activity_results[input_spec.id]
        return _get_item_jsonpath(path, dependency_result)
    if isinstance(input_spec, Object):
        input_ = {}
        for key, subspec in input_spec.items.items():
            value = _build_activity_input(subspec, workflow_input, activity_results)
            if value is not _sentinel:
                input_[key] = value
        return input_
    else:
        raise TypeError(input_spec)


class DAGBuilder(_base.DecisionsBuilder):
    """SWF decision builder from DAG-type workflow specification."""

    def __init__(self, workflow: "DAGWorkflow", task):
        super().__init__(workflow, task)
        self.workflow = workflow
        self._scheduled = {}
        self._activity_task_events = {at.id: [] for at in workflow.task_specs}
        self._new_events = None
        self._error_events = []
        self._ready_activities = set()

    def _schedule_task(self, activity_task: Task):
        workflow_started_event = self.task["events"][0]
        assert workflow_started_event["eventType"] == "WorkflowExecutionStarted"
        attrs = workflow_started_event["workflowExecutionStartedEventAttributes"]
        decision_attributes = {
            "activityId": activity_task.id,
            "activityType": activity_task.type,
        }

        # Build input
        input_spec = activity_task.input
        workflow_input = json.loads(attrs.get("input", "null"))
        activity_results = {}
        for activity_task_id, events in self._activity_task_events.items():
            if activity_task_id in (activity_task.dependencies or []):
                assert events[-1]["eventType"] == "ActivityTaskCompleted"
            elif not events or events[-1]["eventType"] != "ActivityTaskCompleted":
                continue
            d_attrs = events[-1].get("activityTaskCompletedEventAttributes", {})
            if "result" in d_attrs:
                activity_results[activity_task_id] = json.loads(d_attrs["result"])
        input_ = _build_activity_input(input_spec, workflow_input, activity_results)
        if input_ is not _sentinel:
            decision_attributes["input"] = json.dumps(input_)

        # Set other attributes
        if activity_task.heartbeat is not None:
            decision_attributes["heartbeatTimeout"] = str(activity_task.heartbeat)
        if activity_task.timeout is not None:
            decision_attributes["startToCloseTimeout"] = str(activity_task.timeout)
        if activity_task.task_list is not None:
            decision_attributes["taskList"] = {"name": activity_task.task_list}
        if activity_task.priority is not None:
            decision_attributes["taskPriority"] = str(activity_task.priority)

        decision = {
            "decisionType": "ScheduleActivityTask",
            "scheduleActivityTaskDecisionAttributes": decision_attributes,
        }
        self.decisions.append(decision)

    def _get_scheduled_references(self):
        for event in self.task["events"]:
            if event["eventType"] in _activity_events:
                if event["eventType"] == "ActivityTaskScheduled":
                    self._scheduled[event["eventId"]] = event
                else:
                    attrs = event[_attr_keys[event["eventType"]]]
                    self._scheduled[event["eventId"]] = _get(
                        attrs["scheduledEventId"], self.task["events"], "eventId"
                    )

    def _get_activity_task_events(self):
        for event in self.task["events"]:
            if event["eventType"] in _activity_events:
                scheduled_event = self._scheduled[event["eventId"]]
                attrs = scheduled_event["activityTaskScheduledEventAttributes"]
                self._activity_task_events[attrs["activityId"]].append(event)

    def _process_activity_task_completed_event(self, event: t.Dict[str, t.Any]):
        scheduled_event = self._scheduled[event["eventId"]]
        attrs = scheduled_event["activityTaskScheduledEventAttributes"]
        dependants_task = self.workflow.dependants[attrs["activityId"]]

        for activity_task_id in dependants_task:
            assert not self._activity_task_events[activity_task_id]
            task = next(a for a in self.workflow.task_specs if a.id == activity_task_id)

            dependencies_satisfied = True
            for dependency_activity_task_id in task.dependencies:
                events = self._activity_task_events[dependency_activity_task_id]
                if not events or events[-1]["eventType"] != "ActivityTaskCompleted":
                    dependencies_satisfied = False
                    break
            if dependencies_satisfied:
                self._ready_activities.add(task.id)

    def _complete_workflow(self):
        tasks_complete = True
        for events in self._activity_task_events.values():
            if not events or events[-1]["eventType"] != "ActivityTaskCompleted":
                tasks_complete = False
                break

        if tasks_complete:
            result = {}
            for activity_id, events in self._activity_task_events.items():
                assert events and events[-1]["eventType"] == "ActivityTaskCompleted"
                attrs = events[-1].get("activityTaskCompletedEventAttributes")
                if attrs and "result" in attrs:
                    result[activity_id] = json.loads(attrs["result"])

            decision = {"decisionType": "CompleteWorkflowExecution"}
            if result:
                decision_attrs = {"result": json.dumps(result)}
                decision["completeWorkflowExecutionDecisionAttributes"] = decision_attrs
            self.decisions = [decision]

    def _fail_workflow(self, reason=None, details=None):
        decision_attrs = {}
        if reason:
            decision_attrs["reason"] = reason
        if details:
            decision_attrs["details"] = details
        decision = {"decisionType": "FailWorkflowExecution"}
        if decision_attrs:
            decision["failWorkflowExecutionDecisionAttributes"] = decision_attrs
        self.decisions = [decision]

    def _process_decision_failed(self, event: t.Dict[str, t.Any]) -> bool:
        event_ids = [event["eventId"] for event in self.task["events"]]
        attrs = event[_attr_keys[event["eventType"]]]
        if attrs["cause"] == "OPERATION_NOT_PERMITTED":
            idx = event_ids.index(attrs["DecisionTaskCompletedEventId"])
            dc_event = self.task["events"][idx]
            dc_attrs = dc_event["decisionTaskCompletedEventAttributes"]
            idx = event_ids.index(dc_attrs["startedEventId"])
            ds_event = self.task["events"][idx]
            ds_attrs = ds_event["decisionTaskStartedEventAttributes"]
            this_ds_event = self.task["events"][-1]
            this_ds_attrs = this_ds_event["decisionTaskStartedEventAttributes"]
            if ds_attrs["identity"] == this_ds_attrs["identity"]:
                raise _base.DeciderError("Not permitted")
            else:
                return False
        elif attrs["cause"] != "UNHANDLED_DECISION":
            raise _base.DeciderError()

        if event["eventType"] == "CancelWorkflowExecutionFailed":
            self.decisions = [{"decisionType": "CancelWorkflowExecution"}]
            return True
        elif event["eventType"] == "FailWorkflowExecutionFailed":
            return False
        elif event["eventType"] == "CompleteWorkflowExecutionFailed":
            self._complete_workflow()
            return True

    def _schedule_initial_activity_tasks(self):
        for task_id in self.workflow.dependants[None]:
            self._ready_activities.add(task_id)

    def _process_error_events(self):
        if not self._error_events:
            return False
        activity_events = []
        decider_events = []
        time_out_events = []
        other_events = []
        for event in self._error_events:
            if event["eventType"] == "ActivityTaskFailed":
                activity_events.append(event)
            elif event["eventType"] == "ActivityTaskTimedOut":
                attr = event["activityTaskTimedOutEventAttributes"]
                if attr["timeoutType"] in ("START_TO_CLOSE", "HEARTBEAT"):
                    activity_events.append(event)
                elif attr["timeoutType"] in ("SCHEDULE_TO_START", "SCHEDULE_TO_CLOSE"):
                    time_out_events.append(event)
            elif event["eventType"] == "WorkflowExecutionCancelRequested":
                self.decisions = [{"decisionType": "CancelWorkflowExecution"}]
                return True
            elif event["eventType"] in _decision_failed_events:
                if self._process_decision_failed(event):
                    return True
                decider_events.append(event)
            elif event["eventType"] in (
                "DecisionTaskTimedOut",
                "WorkflowExecutionTimedOut",
            ):
                time_out_events.append(event)
            elif event["eventType"] == "RecordMarkerFailed":
                other_events.append(event)

        details = []
        if activity_events:
            details.append("%d activities failed" % len(activity_events))
        if decider_events:
            details.append("%d decisions failed" % len(decider_events))
        if time_out_events:
            details.append("%d actions timed-out" % len(time_out_events))
        if other_events:
            details.append("%d other actions failed" % len(other_events))
        details = ", ".join(details)
        self._fail_workflow(details=details)
        return True

    def _process_event(self, event: t.Dict[str, t.Any]):
        if event["eventType"] == "ActivityTaskCompleted":
            self._process_activity_task_completed_event(event)
        elif event["eventType"] == "WorkflowExecutionStarted":
            self._schedule_initial_activity_tasks()

    def _get_new_events(self):
        event_ids = [event["eventId"] for event in self.task["events"]]
        current_idx = event_ids.index(self.task["startedEventId"])
        previous_idx = -1
        if self.task["previousStartedEventId"] in event_ids:
            previous_idx = event_ids.index(self.task["previousStartedEventId"])
        events = self.task["events"][previous_idx + 1 : current_idx + 1]
        logger.debug(
            "Processing %d events from index %d (ID: %s) to %d (ID: %s)",
            len(events),
            previous_idx + 1,
            events[0]["eventId"],
            current_idx,
            events[-1]["eventId"],
        )
        self._new_events = events

    def _schedule_tasks(self):
        for task_id in self._ready_activities:
            task = next(ts for ts in self.workflow.task_specs if ts.id == task_id)
            assert not self._activity_task_events[task.id]
            self._schedule_task(task)

    def _process_new_events(self):
        assert self.task["events"][-1]["eventType"] == "DecisionTaskStarted"
        assert self.task["events"][-2]["eventType"] == "DecisionTaskScheduled"

        for event in self._new_events[:-2]:
            if event["eventType"] in _error_events:
                self._error_events.append(event)
        if self._process_error_events():
            return

        for event in self._new_events[:-2]:
            self._process_event(event)
        self._schedule_tasks()
        self._complete_workflow()

    def build_decisions(self):
        self._get_scheduled_references()
        self._get_activity_task_events()
        self._get_new_events()
        self._process_new_events()


class DAGWorkflow(_base.Workflow):
    """Dag-type SWF workflow specification.

    Args:
        name: workflow name
        version: workflow version
        task_specs: DAG task specifications
    """

    spec_type = "dag"
    decisions_builder = DAGBuilder
    _task_cls = Task

    def __init__(self, name, version, task_specs: t.List[Task], description=None):
        super().__init__(name, version, description)
        self.task_specs = task_specs
        self.dependants = {None: []}

    @classmethod
    def _args_from_spec(cls, spec):
        args, kwargs = super()._args_from_spec(spec)
        tasks = [cls._task_cls.from_spec(s) for s in spec["tasks"]]
        args += (tasks,)
        return args, kwargs

    def _build_dependants(self):
        for activity_task in self.task_specs:
            dependants_task = []
            for other_activity_task in self.task_specs:
                if activity_task.id in (other_activity_task.dependencies or []):
                    dependants_task.append(other_activity_task.id)
            self.dependants[activity_task.id] = dependants_task
            if not activity_task.dependencies:
                self.dependants[None].append(activity_task.id)

    def setup(self):
        self._build_dependants()



"""Test ``seddy._specs``."""

from unittest import mock

from seddy import _specs as seddy_decisions
import pytest


def test_workflow_map():
    """Test workflow class map."""
    assert seddy_decisions.WORKFLOW["dag"] == seddy_decisions.DAGWorkflow


class TestDecisionsBuilder:
    """Test ``seddy._specs.DecisionsBuilder``."""

    class DecisionsBuilder(seddy_decisions.DecisionsBuilder):
        def build_decisions(self):
            self.decisions = [{"decisionType": "spam"}]

    @pytest.fixture
    def workflow_mock(self):
        """Workflow specification mock."""
        return mock.Mock(spec=seddy_decisions.Workflow)

    @pytest.fixture
    def task(self):
        """Example decision task."""
        return {"taskToken": ""}

    @pytest.fixture
    def instance(self, workflow_mock, task):
        """Decisions builder instance."""
        return self.DecisionsBuilder(workflow_mock, task)

    def test_init(self, instance, workflow_mock, task):
        """Test decisions builder initialisation."""
        assert instance.workflow is workflow_mock
        assert instance.task is task
        assert not instance.decisions


class TestWorkflow:
    """Test ``seddy._specs.Workflow``."""

    class Workflow(seddy_decisions.Workflow):
        class DecisionsBuilder(seddy_decisions.DecisionsBuilder):
            def build_decisions(self):
                self.decisions = [{"decisionType": "spam"}]

        decisions_builder = DecisionsBuilder
        spec_type = "eggs"

    @pytest.fixture
    def spec(self):
        """Example workflow specification."""
        return {
            "name": "foo",
            "version": "0.42",
            "description": "Sunny-side up workflow.",
            "spec_type": "eggs",
            "registration": {
                "active": False,
                "task_timeout": "NONE",
                "execution_timeout": 60,
                "task_list": "spam",
                "task_priority": 2,
                "child_policy": "TERMINATE",
                "lambda_role": "arn:aws:iam::spam:role/eggs",
            },
        }

    @pytest.fixture
    def instance(self):
        """Workflow specification instance."""
        return self.Workflow("foo", "0.42", "Sunny-side up workflow.")

    def test_init(self, instance, spec):
        """Test workflow specification initialisation."""
        assert instance.name == "foo"
        assert instance.version == "0.42"
        assert instance.description == "Sunny-side up workflow."
        assert instance.spec_type == "eggs"
        assert instance.decisions_builder is self.Workflow.DecisionsBuilder

    def test_from_spec(self, spec):
        """Test workflow specification construction from specification."""
        res = self.Workflow.from_spec(spec)
        assert isinstance(res, self.Workflow)
        assert res.name == "foo"
        assert res.version == "0.42"
        assert res.description == "Sunny-side up workflow."
        assert res.registration.active is False
        assert res.registration.task_timeout == "NONE"
        assert res.registration.execution_timeout == 60
        assert res.registration.task_list == "spam"
        assert res.registration.task_priority == 2
        assert res.registration.child_policy == seddy_decisions.ChildPolicy.TERMINATE
        assert res.registration.lambda_role == "arn:aws:iam::spam:role/eggs"

    def test_setup(self, instance):
        """Test workflow specification pre-computation."""
        instance.setup()

    def test_make_decisions(self, instance):
        """Test workflow decision making."""
        task = {"taskToken": ""}
        res = instance.make_decisions(task)
        assert res == [{"decisionType": "spam"}]



"""SWF workflow registration."""

import pathlib
import typing as t
import logging as lg

from . import _util
from . import _specs

logger = lg.getLogger(__name__)


def list_workflows(domain: str, client) -> t.Dict[t.Tuple[str, str], bool]:
    """List all workflows in SWF, including registered and deprecated.

    Args:
        domain: domain to list workflows of
        client (botocore.client.BaseClient): SWF client

    Returns:
        names, versions and registration status of workflows in SWF
    """

    logger.info("Listing workflows in '%s'", domain)

    # List registered workflows
    _kwargs = {"domain": domain, "registrationStatus": "REGISTERED"}
    resp_registered = _util.list_paginated(
        client.list_workflow_types, "typeInfos", _kwargs
    )

    # List deprecated workflows
    _kwargs = {"domain": domain, "registrationStatus": "DEPRECATED"}
    resp_deprecated = _util.list_paginated(
        client.list_workflow_types, "typeInfos", _kwargs
    )

    # Combine
    registered = [w["workflowType"] for w in resp_registered["typeInfos"]]
    deprecated = [w["workflowType"] for w in resp_deprecated["typeInfos"]]
    existing = {(w["name"], w["version"]): True for w in registered}
    existing.update({(w["name"], w["version"]): False for w in deprecated})
    return existing


def register_workflow(workflow: _specs.Workflow, domain: str, client):
    """Register a workflow with SWF.

    Args:
        workflow: specification of workflow to register
        domain: domain to register workflow in
        client (botocore.client.BaseClient): SWF client
    """

    _fmt = "Registering workflow '%s' (version %s) on domain '%s'"
    logger.info(_fmt, workflow.name, workflow.version, domain)

    # Get registration options
    kwargs = {}
    if workflow.description is not None:
        kwargs["description"] = workflow.description
    if workflow.registration:
        if workflow.registration.task_timeout is not None:
            _default = workflow.registration.task_timeout
            kwargs["defaultTaskStartToCloseTimeout"] = str(_default)
        if workflow.registration.execution_timeout is not None:
            _default = workflow.registration.execution_timeout
            kwargs["defaultExecutionStartToCloseTimeout"] = str(_default)
        if workflow.registration.task_list is not None:
            kwargs["defaultTaskList"] = {"name": workflow.registration.task_list}
        if workflow.registration.task_priority is not None:
            kwargs["defaultTaskPriority"] = str(workflow.registration.task_priority)
        if workflow.registration.child_policy is not None:
            kwargs["defaultChildPolicy"] = workflow.registration.child_policy.value
        if workflow.registration.lambda_role is not None:
            kwargs["defaultLambdaRole"] = workflow.registration.lambda_role

    # Register
    client.register_workflow_type(
        domain=domain,
        name=workflow.name,
        version=workflow.version,
        **kwargs,
    )


def deprecate_workflow(workflow: _specs.Workflow, domain: str, client):
    """Deprecate a workflow in SWF.

    Args:
        workflow: specification of workflow to deprecate
        domain: domain to deprecate workflow in
        client (botocore.client.BaseClient): SWF client
    """

    _fmt = "Deprecating workflow '%s' (version %s) in domain '%s'"
    logger.info(_fmt, workflow.name, workflow.version, domain)
    workflow_type = {"name": workflow.name, "version": workflow.version}
    client.deprecate_workflow_type(domain=domain, workflowType=workflow_type)


def undeprecate_workflow(workflow: _specs.Workflow, domain: str, client):
    """Undeprecate a workflow in SWF.

    Args:
        workflow: specification of workflow to undeprecate
        domain: domain to undeprecate workflow in
        client (botocore.client.BaseClient): SWF client
    """

    _fmt = "Undeprecating workflow '%s' (version %s) in domain '%s'"
    logger.info(_fmt, workflow.name, workflow.version, domain)
    workflow_type = {"name": workflow.name, "version": workflow.version}
    client.undeprecate_workflow_type(domain=domain, workflowType=workflow_type)


def _sync_workflow(
    workflow: _specs.Workflow,
    domain: str,
    existing: t.Dict[t.Tuple[str, str], bool],
    client,
):
    """Synchronise a workflow's registration with SWF.

    Args:
        workflow: specification of workflow to register
        domain: domain to register workflow in
        existing:
        client (botocore.client.BaseClient): SWF client
    """

    is_active = workflow.registration.active if workflow.registration else True
    key = (workflow.name, workflow.version)
    if key in existing:
        if existing[key] is is_active:
            _fmt = "Skipping up-to-date workflow '%s' (version %s, active: %s)"
            logger.debug(_fmt, workflow.name, workflow.version, is_active)
        elif is_active:
            undeprecate_workflow(workflow, domain, client)
        elif not is_active:
            deprecate_workflow(workflow, domain, client)
    elif is_active:  # don't register inactive workflows
        register_workflow(workflow, domain, client)


def register_workflows(workflows: t.List[_specs.Workflow], domain: str):
    """Synchronise workflow registration with SWF.

    Args:
        workflows: specifications of workflows to register
        domain: domain to register workflows in
    """

    client = _util.get_swf_client()
    logger.log(25, "Registering workflows in '%s'", domain)

    # Get existing workflows
    existing = list_workflows(domain, client)
    logger.debug("Exising workflows: %s", existing)

    # Register workflows
    for workflow in workflows:
        _sync_workflow(workflow, domain, existing, client)


def run_app(workflows_spec_file: pathlib.Path, domain: str):
    """Run registration synchronisation application.

    Arguments:
        workflows_spec_file: workflows specifications file path
        domain: SWF domain
    """

    workflows = _specs.load_workflows(workflows_spec_file)
    register_workflows(workflows, domain)



"""SWF decider."""

import uuid
import socket
import pathlib
import typing as t
import logging as lg
from concurrent import futures as cf

from . import _util
from . import _specs

logger = lg.getLogger(__name__)
socket.setdefaulttimeout(70.0)


class UnsupportedWorkflow(LookupError):
    """Decider doesn't support workflow."""


class Decider:
    """SWF decider.

    Args:
        workflows_spec_file: workflows specifications file path
        domain: SWF domain to poll in
        task_list: SWF decider task-list
        identity: decider identity, default: automatically generated from
            fully-qualified domain-name and a UUID

    Attributes:
        client (botocore.client.BaseClient): SWF client
        identity (str): name of decider to poll as
    """

    def __init__(
        self,
        workflows_spec_file: pathlib.Path,
        domain: str,
        task_list: str,
        identity: str = None,
    ):
        self.workflows_spec_file = workflows_spec_file
        self.domain = domain
        self.task_list = task_list
        self.client = _util.get_swf_client()
        self.identity = identity or (socket.getfqdn() + "-" + str(uuid.uuid4())[:8])
        self._future = None

    def _poll_for_decision_task(self) -> t.Dict[str, t.Any]:
        """Poll for a decision task from SWF.

        See https://docs.aws.amazon.com/amazonswf/latest/apireference/API_PollForDecisionTask.html

        Returns:
            decision task
        """

        _kwargs = {
            "domain": self.domain,
            "identity": self.identity,
            "taskList": {"name": self.task_list},
        }
        return _util.list_paginated(
            self.client.poll_for_decision_task, "events", _kwargs
        )

    def _get_workflow(self, task: t.Dict[str, t.Any]) -> _specs.Workflow:
        """Get workflow specification for task.

        Args:
            task: decision task

        Returns:
            workflow specification
        """

        workflows = _specs.load_workflows(self.workflows_spec_file)
        workflow_ids = [(w.name, w.version) for w in workflows]
        task_id = (task["workflowType"]["name"], task["workflowType"]["version"])
        try:
            idx = workflow_ids.index(task_id)
        except ValueError:
            raise UnsupportedWorkflow(task["workflowType"]) from None
        return workflows[idx]

    def _respond_decision_task_completed(
        self, decisions: t.List[t.Dict[str, t.Any]], task: t.Dict[str, t.Any]
    ):
        """Send decisions to SWF.

        See https://docs.aws.amazon.com/amazonswf/latest/apireference/API_RespondDecisionTaskCompleted.html

        Args:
            decisions: workflow decisions
            task: decision task
        """

        logger.debug(
            "Sending %d decisions for task '%s'", len(decisions), task["taskToken"]
        )
        self.client.respond_decision_task_completed(
            taskToken=task["taskToken"], decisions=decisions
        )

    def _poll_and_run(self):
        """Perform poll, and possibly run decision task."""
        task = self._poll_for_decision_task()
        logger.debug("Decision task: %s", task)
        if not task["taskToken"]:
            return
        executor = cf.ThreadPoolExecutor(max_workers=1)
        self._future = executor.submit(self._decide_and_respond, task)
        self._future.result()

    def _decide_and_respond(self, task):
        """Make and respond with decisions."""
        logger.info(
            "Got decision task '%s' for workflow '%s-%s' execution '%s' (run '%s')",
            task["taskToken"],
            task["workflowType"]["name"],
            task["workflowType"]["version"],
            task["workflowExecution"]["workflowId"],
            task["workflowExecution"]["runId"],
        )
        try:
            workflow = self._get_workflow(task)
        except UnsupportedWorkflow:
            logger.error("Unsupported workflow type: %s" % task["workflowType"])
            raise
        workflow.setup()

        exc = None
        try:
            decisions = workflow.make_decisions(task)
        except Exception as e:
            decisions = _specs.make_decisions_on_error(e)
            exc = e
        self._respond_decision_task_completed(decisions, task)
        if exc:
            raise exc

    def _run_uncaught(self):
        """Run decider."""
        _fmt = "Polling for tasks in domain '%s' with task-list '%s' as '%s'"
        logger.log(25, _fmt, self.domain, self.task_list, self.identity)
        while True:
            self._poll_and_run()

    def run(self):
        """Run decider."""
        try:
            self._run_uncaught()
        except KeyboardInterrupt:
            logger.info("Quitting due to keyboard-interrupt")
        if self._future and self._future.running():
            logger.log(25, "Waiting on current decision task to be handled")
            self._future.result()


def run_app(
    workflows_spec_file: pathlib.Path, domain: str, task_list: str, identity: str = None
):
    """Run decider application.

    Arguments:
        workflows_spec_file: workflows specifications file path
        domain: SWF domain
        task_list: SWF decider task-list
        identity: decider identity, default: automatically generated
    """

    decider = Decider(workflows_spec_file, domain, task_list, identity)
    decider.run()



"""SWF decider service utilities."""

import os
import sys
import typing as t
import logging as lg

logger = lg.getLogger(__package__)
AWS_SWF_ENDPOINT_URL = os.environ.get("AWS_SWF_ENDPOINT_URL")
LOGGING_LEVELS = {
    -2: lg.ERROR,
    -1: lg.WARNING,
    0: 25,
    1: lg.INFO,
    2: lg.DEBUG,
}


def setup_logging(verbose: int, json_logging: bool = False):
    """Setup logging.

    Args:
        verbose: logging verbosity
        json_logging: JSON-format logs
    """

    lg.addLevelName(25, "NOTICE")
    level = LOGGING_LEVELS.get(verbose, lg.CRITICAL if verbose < 0 else lg.NOTSET)
    fmt = "%(asctime)s [%(levelname)8s] %(name)s: %(message)s"

    if json_logging:
        from pythonjsonlogger import jsonlogger

        handler = lg.StreamHandler()
        formatter = jsonlogger.JsonFormatter(
            "%(levelname)s %(name)s %(message)s", timestamp=True
        )
        handler.setFormatter(formatter)
        lg.basicConfig(level=level, handlers=[handler])
        return

    if level > lg.DEBUG:
        sys.tracebacklimit = 0

    try:
        import coloredlogs
    except ImportError:
        lg.basicConfig(level=level, format=fmt)
        return

    field_styles = {
        "asctime": {"faint": True, "color": "white"},
        "levelname": {"bold": True, "color": "blue"},
        "name": {"bold": True, "color": "yellow"},
    }
    level_styles = {
        **coloredlogs.DEFAULT_LEVEL_STYLES,
        "notice": {},
        "info": {"color": "white"},
    }
    coloredlogs.install(
        level=level,
        fmt=fmt,
        field_styles=field_styles,
        level_styles=level_styles,
        milliseconds=True,
    )
    lg.root.setLevel(level)


def list_paginated(
    fn: t.Callable[..., t.Dict[str, t.Any]],
    list_key: str,
    kwargs: t.Dict[str, t.Any] = None,
    next_key: str = "nextPageToken",
    next_arg: str = None,
) -> t.Dict[str, t.Any]:
    """List AWS resources, consuming pagination.

    Args:
        fn: resource listing function
        list_key: key of paginated list in response
        kwargs: keyword arguments to ``fn``
        next_key: key of next-page token in response
        next_arg: argument name of next-page token in ``fn``, default: same
            as ``next_key``

    Returns:
        collected response of ``fn``
    """

    next_arg = next_key if next_arg is None else next_arg
    kwargs = kwargs or {}
    resp = fn(**kwargs)
    if resp.get(next_key):
        kwargs = kwargs.copy()
        kwargs[next_arg] = resp.pop(next_key)
        new_resp = list_paginated(fn, list_key, kwargs, next_key, next_arg)
        resp[list_key].extend(new_resp[list_key])
    return resp


def get_swf_client():
    """Create an SWF client.

    Uses ``AWS_SWF_ENDPOINT_URL`` from environment for the endpoint URL.

    Returns:
        botocore.client.BaseClient: SWF client
    """

    import boto3

    logger.debug(
        "Creating SWF client with endpoint URL: %s", AWS_SWF_ENDPOINT_URL or "<default>"
    )
    return boto3.client("swf", endpoint_url=AWS_SWF_ENDPOINT_URL)



"""Test ``seddy._util``."""

import sys
import json
from unittest import mock

from seddy import _util as seddy_util
from seddy._specs import _io as seddy_specs_io
import pytest
import yaml


def test_list_paginated():
    # Build input
    def fn(foo, bar=42, nextPageToken=None):
        spams = {None: [0], "spam": [1, 2, 3], "eggs": [4, 7, 9], "ham": [10, 42, 99]}
        tokens = {None: "spam", "spam": "eggs", "eggs": "ham"}
        resp = {"foo": foo * bar, "spam": spams[nextPageToken]}
        if nextPageToken in tokens:
            resp["nextPageToken"] = tokens[nextPageToken]
        return resp

    kwargs = {"foo": "ab", "bar": 7}

    # Run function
    assert seddy_util.list_paginated(fn, "spam", kwargs) == {
        "foo": "ababababababab",
        "spam": [0, 1, 2, 3, 4, 7, 9, 10, 42, 99],
    }


@pytest.fixture
def workflows_spec():
    """Example workflows specifications."""
    return {
        "version": "1.0",
        "workflows": [
            {
                "spec_type": "dag",
                "name": "spam",
                "version": "1.0",
                "tasks": [
                    {
                        "id": "foo",
                        "type": {"name": "spam-foo", "version": "0.3"},
                        "heartbeat": "60",
                        "timeout": "86400",
                        "task_list": "eggs",
                        "priority": "1",
                    }
                ],
            }
        ],
    }


def test_load_workflows_json(tmp_path, workflows_spec):
    """Test workflows specs loading from JSON."""
    # Build input
    workflows_file = tmp_path / "workflows.json"
    workflows_file.write_text(json.dumps(workflows_spec))

    # Run function
    assert seddy_specs_io._load_specs(workflows_file) == workflows_spec


def test_load_workflows_yaml(tmp_path, workflows_spec):
    """Test workflows specs loading from YAML."""
    # Build input
    workflows_file = tmp_path / "workflows.yml"
    workflows_file.write_text(yaml.safe_dump(workflows_spec))

    # Run function
    assert seddy_specs_io._load_specs(workflows_file) == workflows_spec


def test_load_workflows_yaml_raises(tmp_path, workflows_spec):
    """Test workflows specs loading from YAML raises when unavailable."""
    # Setup environment
    yaml_patch = mock.patch.dict(sys.modules, {"yaml": None, "ruamel.yaml": None})

    # Build input
    workflows_file = tmp_path / "workflows.yml"
    workflows_file.write_text(yaml.safe_dump(workflows_spec))

    # Run function
    with pytest.raises(ModuleNotFoundError), yaml_patch:
        seddy_specs_io._load_specs(workflows_file)


def test_load_workflows_with_incorrect_suffix(tmp_path, workflows_spec):
    """Test workflows specs loading raises for incorrect suffix."""
    # Build input
    workflows_file = tmp_path / "workflows.spam"
    workflows_file.write_text(str(workflows_spec))

    # Run function
    with pytest.raises(ValueError):
        seddy_specs_io._load_specs(workflows_file)



"""Test ``seddy`` command-line application."""

import sys
import json
import logging as lg
from unittest import mock

from seddy import __main__ as seddy_main
from seddy import decider as seddy_decider
from seddy import registration as seddy_registration
import pytest
import coloredlogs
import pkg_resources


@pytest.fixture
def decider_mock():
    """Decider application mock."""
    run_app_mock = mock.Mock()
    with mock.patch.object(seddy_decider, "run_app", run_app_mock):
        yield run_app_mock


@pytest.mark.parametrize(
    ("verbosity_flags", "exp_logging_level"),
    [
        pytest.param([], 25, id='""'),
        pytest.param(["-v"], lg.INFO, id='"-v"'),
        pytest.param(["-vv"], lg.DEBUG, id='"-vv"'),
        pytest.param(["-vvv"], lg.NOTSET, id='"-vvv"'),
        pytest.param(["-vvvv"], lg.NOTSET, id='"-vvvv"'),
        pytest.param(["-q"], lg.WARNING, id='"-q"'),
        pytest.param(["-qq"], lg.ERROR, id='"-qq"'),
        pytest.param(["-qqq"], lg.CRITICAL, id='"-qqq"'),
        pytest.param(["-qqqq"], lg.CRITICAL, id='"-qqqq"'),
        pytest.param(["-vq"], 25, id='"-vq"'),
        pytest.param(["-vvqq"], 25, id='"-vvqq"'),
        pytest.param(["-vvq"], lg.INFO, id='"-vvq"'),
        pytest.param(["-vqq"], lg.WARNING, id='"-vqq"'),
        pytest.param(["-v", "-v"], lg.DEBUG, id='"-v -v"'),
        pytest.param(["-v", "-q"], 25, id='"-v -q"'),
        pytest.param(["-q", "-q"], lg.ERROR, id='"-q -q"'),
    ],
)
@pytest.mark.parametrize(
    "coloredlogs_module",
    [pytest.param(None, id="logging"), pytest.param(coloredlogs, id="coloredlogs")],
)
def test_logging(
    decider_mock,
    tmp_path,
    verbosity_flags,
    exp_logging_level,
    coloredlogs_module,
    capsys,
):
    """Ensure logging configuration is set up correctly."""
    # Setup environment
    coloredlogs_patch = mock.patch.dict(
        sys.modules, {"coloredlogs": coloredlogs_module}
    )

    root_logger = lg.RootLogger("WARNING")
    root_logger_patch = mock.patch.object(lg, "root", root_logger)

    # Run function
    parser = seddy_main.build_parser()
    args = parser.parse_args(
        verbosity_flags + ["decider", str(tmp_path / "workflows.json"), "spam", "eggs"]
    )
    with root_logger_patch, coloredlogs_patch:
        seddy_main.run_app(args)

    # Check logging configuration
    assert root_logger.level == exp_logging_level

    root_logger.critical("spam")
    assert capsys.readouterr().err[24:] == "[CRITICAL] root: spam\n"


@pytest.mark.parametrize(
    "coloredlogs_module",
    [
        pytest.param(None, id="no-coloredlogs"),
        pytest.param(coloredlogs, id="coloredlogs"),
    ],
)
def test_json_logging(decider_mock, tmp_path, coloredlogs_module, capsys):
    """Ensure JSON logging configuration is set up correctly."""
    # Setup environment
    root_logger = lg.RootLogger("WARNING")
    root_logger_patch = mock.patch.object(lg, "root", root_logger)

    # Run function
    parser = seddy_main.build_parser()
    args = parser.parse_args(
        ["-J", "decider", str(tmp_path / "workflows.json"), "spam", "eggs"]
    )
    with root_logger_patch:
        seddy_main.run_app(args)

    # Check logging configuration
    root_logger.warning("spam %s", "eggs")
    assert json.loads(capsys.readouterr().err) == {
        "levelname": "WARNING",
        "name": "root",
        "timestamp": mock.ANY,
        "message": "spam eggs",
    }


@pytest.mark.parametrize(
    ("command_line_args", "description"),
    [
        pytest.param(["-h"], "SWF workflow management service.", id='"-h"'),
        pytest.param(["--help"], "SWF workflow management service.", id='"--help"'),
        pytest.param(["decider", "-h"], "Run SWF decider.", id='"decider -h"'),
        pytest.param(
            ["decider", "a.json", "spam", "eggs", "-h"],
            "Run SWF decider.",
            id='"decider a.json spam eggs -h"',
        ),
        pytest.param(
            ["register", "-h"],
            "Synchronise workflow registration status with SWF.",
            id='"register -h"',
        ),
    ],
)
def test_usage(decider_mock, command_line_args, capsys, description):
    """Ensure usage is displayed."""
    # Run function
    parser = seddy_main.build_parser()
    with pytest.raises(SystemExit) as e:
        parser.parse_args(command_line_args)
    assert e.value.code == 0

    # Check output
    res_out = capsys.readouterr().out
    assert res_out[:6] == "usage:"
    assert res_out.splitlines()[2] == description


@pytest.mark.parametrize(
    "command_line_args",
    [
        pytest.param(["-V"], id='"-V"'),
        pytest.param(["--version"], id='"--version"'),
        pytest.param(
            ["-V", "decider", "a.json", "spam", "eggs"],
            id='"-V decider a.json spam eggs"',
        ),
    ],
)
def test_version(decider_mock, command_line_args, capsys):
    """Ensure version is displayed."""
    # Run function
    parser = seddy_main.build_parser()
    with pytest.raises(SystemExit) as e:
        parser.parse_args(command_line_args)
    assert e.value.code == 0

    # Check output
    res_out = capsys.readouterr().out
    assert res_out.strip() == pkg_resources.get_distribution("seddy").version


@pytest.mark.parametrize(
    ("args_extra", "decider_args"),
    [
        pytest.param([], [None], id='""'),
        pytest.param(["-i", "abcd1234"], ["abcd1234"], id='"-i abcd1234"'),
    ],
)
def test_decider(decider_mock, tmp_path, args_extra, decider_args):
    """Ensure decider application is run with the correct input."""
    # Run function
    parser = seddy_main.build_parser()
    args = parser.parse_args(
        ["decider", str(tmp_path / "workflows.json"), "spam", "eggs"] + args_extra
    )
    seddy_main.run_app(args)

    # Check application input
    decider_mock.assert_called_once_with(
        tmp_path / "workflows.json", "spam", "eggs", *decider_args
    )


def test_register(tmp_path):
    """Ensure workflow registration application is run correctly."""
    # Setup environment
    run_app_mock = mock.Mock()
    run_app_patch = mock.patch.object(seddy_registration, "run_app", run_app_mock)

    # Run function
    parser = seddy_main.build_parser()
    args = parser.parse_args(["register", str(tmp_path / "workflows.json"), "spam"])
    with run_app_patch:
        seddy_main.run_app(args)

    # Check application input
    run_app_mock.assert_called_once_with(tmp_path / "workflows.json", "spam")



"""Multi-workflow SWF decider and workflow management service."""

__all__ = [
    "ChildPolicy",
    "Registration",
    "DecisionsBuilder",
    "Workflow",
    "DAGBuilder",
    "DAGWorkflow",
    "load_workflows",
    "WORKFLOW",
]

from ._specs import ChildPolicy
from ._specs import Registration
from ._specs import DecisionsBuilder
from ._specs import Workflow
from ._specs import DAGBuilder
from ._specs import DAGWorkflow
from ._specs import load_workflows
from ._specs import WORKFLOW



"""SWF workflow management service."""

import pathlib
import argparse

import pkg_resources

try:
    from pythonjsonlogger import jsonlogger
except ImportError as e:  # pragma: no cover
    jsonlogger = e

try:
    version = pkg_resources.get_distribution("seddy").version
except pkg_resources.DistributionNotFound:  # pragma: no cover
    version = None


def run_app(args: argparse.Namespace):
    """Run application from parsed command-line arguments."""
    from . import _util

    _util.setup_logging(args.verbose - args.quiet, args.json_logging)

    if args.command == "decider":
        from . import decider

        decider.run_app(args.workflows_file, args.domain, args.task_list, args.identity)
    elif args.command == "register":
        from . import registration

        registration.run_app(args.workflows_file, args.domain)
    else:  # pragma: no cover
        raise ValueError(args.command)


def build_parser() -> argparse.ArgumentParser:
    """Build command-line argument parser."""
    parser = argparse.ArgumentParser(description=__doc__.splitlines()[0])
    parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="increase logging verbosity"
    )
    parser.add_argument(
        "-q", "--quiet", action="count", default=0, help="decrease logging verbosity"
    )
    if not isinstance(jsonlogger, Exception):
        parser.add_argument(
            "-J",
            "--json-logging",
            action="store_true",
            help="JSON-format logs (coloured-logging disabled)",
        )
    parser.add_argument("-V", "--version", action="version", version=version)
    subparsers = parser.add_subparsers(dest="command", metavar="COMMAND")
    subparsers.required = True

    # Decider
    decider_parser = subparsers.add_parser(
        "decider", help="run SWF decider", description="Run SWF decider."
    )
    decider_parser.add_argument(
        "workflows_file", type=pathlib.Path, help="workflows specifications file path"
    )
    decider_parser.add_argument("domain", help="SWF domain")
    decider_parser.add_argument("task_list", help="SWF decider task-list")
    decider_parser.add_argument(
        "-i",
        "--identity",
        metavar="NAME",
        help="decider identity, default: automatically generated",
    )

    # Workflows registration
    register_parser = subparsers.add_parser(
        "register",
        help="synchronise workflow registration status with SWF",
        description="Synchronise workflow registration status with SWF.",
    )
    register_parser.add_argument(
        "workflows_file", type=pathlib.Path, help="workflows specifications file path"
    )
    register_parser.add_argument("domain", help="SWF domain")

    return parser


def main():  # pragma: no cover
    parser = build_parser()
    args = parser.parse_args()
    run_app(args)


if __name__ == "__main__":  # pragma: no cover
    main()



"""SWF workflows specifications."""

__all__ = [
    "ChildPolicy",
    "Registration",
    "DecisionsBuilder",
    "Workflow",
    "make_decisions_on_error",
    "DAGBuilder",
    "DAGWorkflow",
    "load_workflows",
    "WORKFLOW",
]

from ._base import ChildPolicy
from ._base import Registration
from ._base import DecisionsBuilder
from ._base import Workflow
from ._base import make_decisions_on_error
from ._dag import DAGBuilder
from ._dag import DAGWorkflow
from ._io import load_workflows

WORKFLOW = {
    DAGWorkflow.spec_type: DAGWorkflow,
}



"""Configuration file for the Sphinx documentation builder."""

import pkg_resources

from sphinx.ext import apidoc

apidoc.main(["-eTf", "-t", "../templates", "-o", ".", "../../src"])  # gen API docs

project = "seddy"
copyright = "2020, Laurie O"
author = "Laurie O"

release = pkg_resources.get_distribution("seddy").version  # full version
version = ".".join(release.split(".")[:2])  # short version

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "autodocsumm",
]
html_theme = "sphinx_rtd_theme"
master_doc = "index"  # support read-the-docs



"""Test ``seddy._specs._dag``."""

import logging as lg

from seddy import _specs as seddy_specs
from seddy._specs import _dag
import pytest

lg.root.setLevel(lg.DEBUG)


@pytest.mark.parametrize(
    ("path", "obj", "exp"),
    [
        ("$", None, None),
        ("$", {}, {}),
        ("$", [], []),
        ("$", 42, 42),
        ("$", "spam", "spam"),
        (
            "$.spam[0].eggs.swallow[2]",
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            42,
        ),
        (
            "$.spam[0].eggs.swallow",
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            [None, None, 42],
        ),
        (
            "$.spam[0].eggs",
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            {"swallow": [None, None, 42]},
        ),
        (
            "$.spam[0]",
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            {"eggs": {"swallow": [None, None, 42]}},
        ),
        (
            "$.spam",
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            [{"eggs": {"swallow": [None, None, 42]}}, False],
        ),
        (
            "$",
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
        ),
        ("$[1]", [{"eggs": {"swallow": [None, None, 42]}}, False], False),
    ],
)
def test_get_item_jsonpath(path, obj, exp):
    assert _dag._get_item_jsonpath(path, obj) == exp


@pytest.mark.parametrize(
    "path",
    [
        "$spam[0].eggs.swallow[2]",
        ".spam[0].eggs.swallow[2]",
        "spam[0].eggs.swallow[2]",
        "$.spam0].eggs.swallow[2]",
        "$.spam[0.eggs.swallow[2]",
        "$.spam[[0]].eggs.swallow[2]",
        "$.spam[0.eggs].swallow[2]",
        "$.spam[a].eggs.swallow[2]",
        "$.spam[0][eggs].swallow[2]",
        "$.spam[0].eggs-swallow[2]",
        "$.spam[0].eggs$.swallow[2]",
        "$.spam[0]].eggs.swallow[2]",
        "$.spam[[0].eggs.swallow[2]",
        "$.spam[0].eggs.swallow[*]",
    ],
)
def test_get_item_jsonpath_bad_path(path):
    obj = {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]}
    with pytest.raises(ValueError):
        _dag._get_item_jsonpath(path, obj)


@pytest.mark.parametrize(
    ("spec", "workflow_input", "activity_results", "exp"),
    [
        pytest.param(None, None, {}, _dag._sentinel, id="none"),
        pytest.param(
            None, {"foo": {"spam": 42}, "bar": False}, {}, _dag._sentinel, id="none"
        ),
        pytest.param(_dag.NoInput(), None, {}, _dag._sentinel, id="none"),
        pytest.param(
            _dag.Constant(None), None, {}, None, id="constant"
        ),
        pytest.param(
            _dag.Constant({"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]}),
            None,
            {},
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            id="constant",
        ),
        pytest.param(
            _dag.Constant({"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]}),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {},
            {"spam": [{"eggs": {"swallow": [None, None, 42]}}, False]},
            id="constant",
        ),
        pytest.param(
            _dag.WorkflowInput(),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {},
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            id="workflow-input",
        ),
        pytest.param(
            _dag.WorkflowInput("$.foo"),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {},
            {"spam": 42, "eggs": None, "ham": [1, 2]},
            id="workflow-input",
        ),
        pytest.param(
            _dag.WorkflowInput("$.foo.ham[1]"),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {},
            2,
            id="workflow-input",
        ),
        pytest.param(
            _dag.DependencyResult("foo"),
            None,
            {"foo": [{"spam": "eggs", "ham": 42}, False], "bar": None},
            [{"spam": "eggs", "ham": 42}, False],
            id="dependency-result",
        ),
        pytest.param(
            _dag.DependencyResult("foo"),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {"foo": [{"spam": "eggs", "ham": 42}, False], "bar": None},
            [{"spam": "eggs", "ham": 42}, False],
            id="dependency-result",
        ),
        pytest.param(
            _dag.DependencyResult("foo", "$[0].spam"),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {"foo": [{"spam": "eggs", "ham": 42}, False], "bar": None},
            "eggs",
            id="dependency-result",
        ),
        pytest.param(
            _dag.Object(
                {
                    "pie": _dag.DependencyResult("foo", "$[0].spam"),
                    "cheese": _dag.Object(
                        {
                            "a": _dag.Constant(["b"]),
                            "c": _dag.NoInput(),
                            "d": _dag.DependencyResult("bar"),
                        },
                    ),
                    "gravy": _dag.WorkflowInput("$.bar"),
                },
            ),
            {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False},
            {"foo": [{"spam": "eggs", "ham": 42}, False], "bar": None},
            {"pie": "eggs", "cheese": {"a": ["b"], "d": None}, "gravy": False},
            id="object",
        ),
    ],
)
def test_build_activity_input(spec, workflow_input, activity_results, exp):
    assert _dag._build_activity_input(spec, workflow_input, activity_results) == exp


def test_build_activity_input_unknown_type():
    input_spec = _dag.TaskInput()
    workflow_input = {"foo": {"spam": 42, "eggs": None, "ham": [1, 2]}, "bar": False}
    activity_results = {"foo": [{"spam": "eggs", "ham": 42}, False], "bar": None}
    with pytest.raises(TypeError):
        _dag._build_activity_input(input_spec, workflow_input, activity_results)


class TestDAGDecisionsBuilding:
    """Test ``seddy._specs.DAGBuilder``."""

    @pytest.fixture
    def workflow(self):
        """Example DAG workflow specification."""
        workflow = seddy_specs.DAGWorkflow.from_spec(
            {
                "name": "foo",
                "version": "0.42",
                "tasks": [
                    {
                        "id": "foo",
                        "type": {"name": "spam-foo", "version": "0.3"},
                        "input": {"type": "workflow-input", "path": "$.foo"},
                        "heartbeat": 60,
                        "timeout": 86400,
                        "task_list": "eggs",
                        "priority": 1,
                    },
                    {
                        "id": "bar",
                        "type": {"name": "spam-bar", "version": "0.1"},
                        "input": {"type": "workflow-input", "path": "$.bar"},
                        "heartbeat": 60,
                        "timeout": 86400,
                        "dependencies": ["foo"],
                    },
                    {
                        "id": "yay",
                        "type": {"name": "spam-foo", "version": "0.3"},
                        "input": {"type": "workflow-input", "path": "$.yay"},
                        "heartbeat": 60,
                        "timeout": 86400,
                        "dependencies": ["foo"],
                    },
                    {
                        "id": "tin",
                        "type": {"name": "spam-tin", "version": "1.2"},
                        "input": {"type": "none"},
                        "heartbeat": 30,
                        "timeout": 43200,
                        "dependencies": ["yay"],
                    },
                ],
                "type": "dag",
            }
        )
        workflow.setup()
        return workflow

    def test_workflow_start(self, workflow):
        """Test DAG decisions building on workflow start."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 0,
            "startedEventId": 3,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "foo",
                    "activityType": {"name": "spam-foo", "version": "0.3"},
                    "input": '{"spam": [42], "eggs": null}',
                    "heartbeatTimeout": "60",
                    "startToCloseTimeout": "86400",
                    "taskPriority": "1",
                    "taskList": {"name": "eggs"},
                },
            },
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_other_event(self, workflow):
        """Test DAG decisions building after other event."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 7,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "WorkflowExecutionSignaled",
                    "workflowExecutionSignaledEventAttributes": {"signalName": "blue"},
                },
                {"eventId": 6, "eventType": "DecisionTaskScheduled"},
                {"eventId": 7, "eventType": "DecisionTaskStarted"},
            ],
        }
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == []

    def test_foo_complete(self, workflow):
        """Test DAG decisions building after foo activity completes."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "bar",
                    "activityType": {"name": "spam-bar", "version": "0.1"},
                    "input": "null",
                    "heartbeatTimeout": "60",
                    "startToCloseTimeout": "86400",
                },
            },
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "yay",
                    "activityType": {"name": "spam-foo", "version": "0.3"},
                    "input": '{"spam": [17], "eggs": [42]}',
                    "heartbeatTimeout": "60",
                    "startToCloseTimeout": "86400",
                },
            },
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions in (expected_decisions, expected_decisions[::-1])

    def test_foo_complete_yay_unsatisfied(self, workflow):
        """Test DAG decisions building after foo completes yet yay not ready."""
        workflow.dependants["bar"] = ["yay"]
        assert workflow.task_specs[2].id == "yay"
        workflow.task_specs[2].dependencies = ["foo", "bar"]
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "bar",
                    "activityType": {"name": "spam-bar", "version": "0.1"},
                    "input": "null",
                    "heartbeatTimeout": "60",
                    "startToCloseTimeout": "86400",
                },
            },
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_yay_complete(self, workflow):
        """Test DAG decisions building after yay activity completes."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 9,
            "startedEventId": 17,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
                {"eventId": 10, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 11,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "yay",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 10,
                        "input": '{"spam": [17], "eggs": [42]}',
                    },
                },
                {
                    "eventId": 12,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "bar",
                        "activityType": {"name": "spam-bar", "version": "0.1"},
                        "decisionTaskCompletedEventId": 10,
                        "input": "null",
                    },
                },
                {
                    "eventId": 13,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 11},
                },
                {
                    "eventId": 14,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 12},
                },
                {
                    "eventId": 15,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 11,
                        "result": "5",
                    },
                },
                {"eventId": 16, "eventType": "DecisionTaskScheduled"},
                {"eventId": 17, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "tin",
                    "activityType": {"name": "spam-tin", "version": "1.2"},
                    "heartbeatTimeout": "30",
                    "startToCloseTimeout": "43200",
                },
            },
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_bar_complete(self, workflow):
        """Test DAG decisions building after bar activity completes."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 17,
            "startedEventId": 23,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
                {"eventId": 10, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 11,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "yay",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 10,
                        "input": '{"spam": [17], "eggs": [42]}',
                    },
                },
                {
                    "eventId": 12,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "bar",
                        "activityType": {"name": "spam-bar", "version": "0.1"},
                        "decisionTaskCompletedEventId": 10,
                        "input": "null",
                    },
                },
                {
                    "eventId": 13,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 11},
                },
                {
                    "eventId": 14,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 12},
                },
                {
                    "eventId": 15,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 11,
                        "result": "5",
                    },
                },
                {"eventId": 16, "eventType": "DecisionTaskScheduled"},
                {"eventId": 17, "eventType": "DecisionTaskStarted"},
                {"eventId": 18, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 19,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "tin",
                        "activityType": {"name": "spam-tin", "version": "1.2"},
                        "decisionTaskCompletedEventId": 18,
                    },
                },
                {
                    "eventId": 20,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 19},
                },
                {
                    "eventId": 21,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 12,
                        "result": '{"a": 9, "b": "red"}',
                    },
                },
                {"eventId": 22, "eventType": "DecisionTaskScheduled"},
                {"eventId": 23, "eventType": "DecisionTaskStarted"},
            ],
        }
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == []

    def test_tin_complete(self, workflow):
        """Test DAG decisions building after tin activity completes."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 23,
            "startedEventId": 27,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
                {"eventId": 10, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 11,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "yay",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 10,
                        "input": '{"spam": [17], "eggs": [42]}',
                    },
                },
                {
                    "eventId": 12,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "bar",
                        "activityType": {"name": "spam-bar", "version": "0.1"},
                        "decisionTaskCompletedEventId": 10,
                        "input": "null",
                    },
                },
                {
                    "eventId": 13,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 11},
                },
                {
                    "eventId": 14,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 12},
                },
                {
                    "eventId": 15,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 11,
                        "result": "5",
                    },
                },
                {"eventId": 16, "eventType": "DecisionTaskScheduled"},
                {"eventId": 17, "eventType": "DecisionTaskStarted"},
                {"eventId": 18, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 19,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "tin",
                        "activityType": {"name": "spam-tin", "version": "1.2"},
                        "decisionTaskCompletedEventId": 18,
                    },
                },
                {
                    "eventId": 20,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 19},
                },
                {
                    "eventId": 21,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 12,
                        "result": '{"a": 9, "b": "red"}',
                    },
                },
                {"eventId": 22, "eventType": "DecisionTaskScheduled"},
                {"eventId": 23, "eventType": "DecisionTaskStarted"},
                {"eventId": 24, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 25,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {"scheduledEventId": 19},
                },
                {"eventId": 26, "eventType": "DecisionTaskScheduled"},
                {"eventId": 27, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "CompleteWorkflowExecution",
                "completeWorkflowExecutionDecisionAttributes": {
                    "result": '{"foo": 3, "bar": {"a": 9, "b": "red"}, "yay": 5}'
                },
            }
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_bar_and_yay_complete(self, workflow):
        """Test DAG decisions building after bar and yay activities complete."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 9,
            "startedEventId": 18,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
                {"eventId": 10, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 11,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "yay",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 10,
                        "input": '{"spam": [17], "eggs": [42]}',
                    },
                },
                {
                    "eventId": 12,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "bar",
                        "activityType": {"name": "spam-bar", "version": "0.1"},
                        "decisionTaskCompletedEventId": 10,
                        "input": "null",
                    },
                },
                {
                    "eventId": 13,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 11},
                },
                {
                    "eventId": 14,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 12},
                },
                {
                    "eventId": 15,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 11,
                        "result": "5",
                    },
                },
                {
                    "eventId": 16,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 12,
                        "result": '{"a": 9, "b": "red"}',
                    },
                },
                {"eventId": 17, "eventType": "DecisionTaskScheduled"},
                {"eventId": 18, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "tin",
                    "activityType": {"name": "spam-tin", "version": "1.2"},
                    "heartbeatTimeout": "30",
                    "startToCloseTimeout": "43200",
                },
            },
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_foo_failed(self, workflow):
        """Test DAG decisions building after foo activity fails."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskFailed",
                    "activityTaskFailedEventAttributes": {
                        "scheduledEventId": 5,
                        "details": "The specified spam does not exist",
                        "reason": "spamError",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "FailWorkflowExecution",
                "failWorkflowExecutionDecisionAttributes": {
                    "details": "1 activities failed",
                },
            }
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_foo_timed_out(self, workflow):
        """Test DAG decisions building after foo activity times-out."""
        # Events sections
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskTimedOut",
                    "activityTaskTimedOutEventAttributes": {
                        "scheduledEventId": 5,
                        "details": "42 / 50",
                        "timeoutType": "HEARTBEAT",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "FailWorkflowExecution",
                "failWorkflowExecutionDecisionAttributes": {
                    "details": "1 activities failed",
                },
            }
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_foo_start_timed_out(self, workflow):
        """Test DAG decisions building after foo activity start times-out."""
        # Events sections
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskTimedOut",
                    "activityTaskTimedOutEventAttributes": {
                        "scheduledEventId": 5,
                        "timeoutType": "SCHEDULE_TO_START",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "FailWorkflowExecution",
                "failWorkflowExecutionDecisionAttributes": {
                    "details": "1 actions timed-out",
                },
            }
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_decision_timed_out(self, workflow):
        """Test DAG decisions building after decision task times-out."""
        # Events sections
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {"eventId": 7, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 8,
                    "eventType": "DecisionTaskTimedOut",
                    "workflowExecutionTimedOutTimedOutEventAttributes": {
                        "timeoutType": "START_TO_CLOSE",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [
            {
                "decisionType": "FailWorkflowExecution",
                "failWorkflowExecutionDecisionAttributes": {
                    "details": "1 actions timed-out",
                },
            }
        ]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_workflow_cancel(self, workflow):
        """Test DAG decisions building after workflow is cancelled."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 9,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {"eventId": 7, "eventType": "WorkflowExecutionCancelRequested"},
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [{"decisionType": "CancelWorkflowExecution"}]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    def test_workflow_cancel_after_bar_yay(self, workflow):
        """Test DAG decisions building: workflow cancelled after bar and yay."""
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 9,
            "startedEventId": 16,
            "events": [
                {
                    "eventId": 1,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "input": (
                            "{\n"
                            '    "foo": {"spam": [42], "eggs": null},\n'
                            '    "bar": null,\n'
                            '    "yay": {"spam": [17], "eggs": [42]}\n'
                            "}"
                        )
                    },
                },
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {"eventId": 3, "eventType": "DecisionTaskStarted"},
                {"eventId": 4, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 5,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "foo",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 4,
                        "input": '{"spam": [42], "eggs": null}',
                    },
                },
                {
                    "eventId": 6,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 5},
                },
                {
                    "eventId": 7,
                    "eventType": "ActivityTaskCompleted",
                    "activityTaskCompletedEventAttributes": {
                        "scheduledEventId": 5,
                        "result": "3",
                    },
                },
                {"eventId": 8, "eventType": "DecisionTaskScheduled"},
                {"eventId": 9, "eventType": "DecisionTaskStarted"},
                {"eventId": 10, "eventType": "DecisionTaskCompleted"},
                {
                    "eventId": 11,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "yay",
                        "activityType": {"name": "spam-foo", "version": "0.3"},
                        "decisionTaskCompletedEventId": 10,
                        "input": '{"spam": [17], "eggs": [42]}',
                    },
                },
                {
                    "eventId": 12,
                    "eventType": "ActivityTaskScheduled",
                    "activityTaskScheduledEventAttributes": {
                        "activityId": "bar",
                        "activityType": {"name": "spam-bar", "version": "0.1"},
                        "decisionTaskCompletedEventId": 10,
                        "input": "null",
                    },
                },
                {
                    "eventId": 13,
                    "eventType": "ActivityTaskStarted",
                    "activityTaskStartedEventAttributes": {"scheduledEventId": 11},
                },
                {"eventId": 14, "eventType": "WorkflowExecutionCancelRequested"},
                {"eventId": 15, "eventType": "DecisionTaskScheduled"},
                {"eventId": 16, "eventType": "DecisionTaskStarted"},
            ],
        }
        expected_decisions = [{"decisionType": "CancelWorkflowExecution"}]
        instance = seddy_specs.DAGBuilder(workflow, task)
        instance.build_decisions()
        assert instance.decisions == expected_decisions

    @pytest.mark.parametrize(
        ("cause", "identity", "event_type", "exp"),
        [
            pytest.param(
                "OPERATION_NOT_PERMITTED",
                "spam-1235",
                "ScheduleActivityTaskFailed",
                [
                    {
                        "decisionType": "FailWorkflowExecution",
                        "failWorkflowExecutionDecisionAttributes": {
                            "details": "1 decisions failed",
                        },
                    }
                ],
                id="permissions-other",
            ),
            pytest.param(
                "UNHANDLED_DECISION",
                "spam-1234",
                "CancelWorkflowExecutionFailed",
                [{"decisionType": "CancelWorkflowExecution"}],
                id="unhandled-cancel",
            ),
            pytest.param(
                "UNHANDLED_DECISION",
                "spam-1234",
                "FailWorkflowExecutionFailed",
                [
                    {
                        "decisionType": "FailWorkflowExecution",
                        "failWorkflowExecutionDecisionAttributes": {
                            "details": "1 decisions failed",
                        },
                    }
                ],
                id="unhandled-fail",
            ),
            pytest.param(
                "UNHANDLED_DECISION",
                "spam-1234",
                "CompleteWorkflowExecutionFailed",
                [],
                id="unhandled-complete",
            ),
        ],
    )
    def test_decision_failure(self, workflow, cause, identity, event_type, exp):
        """Test decision failure handling."""
        # Build input
        event_attr_key = event_type[0].lower() + event_type[1:] + "EventAttributes"
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 7,
            "events": [
                {"eventId": 1, "eventType": "WorkflowExecutionStarted"},
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 3,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {"identity": "spam-1234"},
                },
                {
                    "eventId": 4,
                    "eventType": "DecisionTaskCompleted",
                    "decisionTaskCompletedEventAttributes": {"startedEventId": 3},
                },
                {
                    "eventId": 5,
                    "eventType": event_type,
                    event_attr_key: {"cause": cause, "DecisionTaskCompletedEventId": 4},
                },
                {"eventId": 6, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 7,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {"identity": identity},
                },
            ],
        }
        instance = seddy_specs.DAGBuilder(workflow, task)

        # Run function
        instance.build_decisions()
        assert instance.decisions == exp

    @pytest.mark.parametrize(
        ("cause", "exception_str"),
        [
            pytest.param("OPERATION_NOT_PERMITTED", "Not permitted", id="permissions"),
            pytest.param("INVALID_INPUT", "", id="error"),
        ],
    )
    def test_decision_failure_raises(self, workflow, cause, exception_str):
        """Test decision failure handling raises on configuration issue."""
        # Build input
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 7,
            "events": [
                {"eventId": 1, "eventType": "WorkflowExecutionStarted"},
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 3,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {"identity": "spam-1234"},
                },
                {
                    "eventId": 4,
                    "eventType": "DecisionTaskCompleted",
                    "decisionTaskCompletedEventAttributes": {"startedEventId": 3},
                },
                {
                    "eventId": 5,
                    "eventType": "CancelWorkflowExecutionFailed",
                    "cancelWorkflowExecutionFailedEventAttributes": {
                        "cause": cause,
                        "DecisionTaskCompletedEventId": 4,
                    },
                },
                {"eventId": 6, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 7,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {"identity": "spam-1234"},
                },
            ],
        }
        instance = seddy_specs.DAGBuilder(workflow, task)

        # Run function
        with pytest.raises(seddy_specs._base.DeciderError) as e:
            instance.build_decisions()
        assert str(e.value) == exception_str

    def test_workflow_complete_decision_failure(self):
        """Test workflow-complete decision failure handling."""
        # Build input
        workflow = seddy_specs.DAGWorkflow.from_spec(
            {"name": "foo", "version": "0.44", "tasks": [], "type": "dag"}
        )
        workflow.setup()
        task = {
            "taskToken": "spam",
            "previousStartedEventId": 3,
            "startedEventId": 7,
            "events": [
                {"eventId": 1, "eventType": "WorkflowExecutionStarted"},
                {"eventId": 2, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 3,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {"identity": "spam-1234"},
                },
                {
                    "eventId": 4,
                    "eventType": "DecisionTaskCompleted",
                    "decisionTaskCompletedEventAttributes": {"startedEventId": 3},
                },
                {
                    "eventId": 5,
                    "eventType": "CompleteWorkflowExecutionFailed",
                    "completeWorkflowExecutionFailedEventAttributes": {
                        "cause": "UNHANDLED_DECISION",
                        "DecisionTaskCompletedEventId": 4,
                    },
                },
                {"eventId": 6, "eventType": "DecisionTaskScheduled"},
                {
                    "eventId": 7,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {"identity": "spam-1234"},
                },
            ],
        }
        instance = seddy_specs.DAGBuilder(workflow, task)

        # Run function
        instance.build_decisions()
        assert instance.decisions == [{"decisionType": "CompleteWorkflowExecution"}]


class TestWorkflow:
    """Test ``seddy._specs.DAGWorkflow``."""

    @pytest.fixture
    def task_spec_dicts(self):
        """Example DAG-type workflow tasks specifications."""
        return [
            {
                "id": "foo",
                "type": {"name": "spam-foo", "version": "0.3"},
                "heartbeat": "60",
                "timeout": "86400",
                "task_list": "eggs",
                "priority": "1",
            },
            {
                "id": "bar",
                "type": {"name": "spam-bar", "version": "0.1"},
                "heartbeat": "60",
                "timeout": "86400",
                "dependencies": "foo",
            },
            {
                "id": "yay",
                "type": {"name": "spam-foo", "version": "0.3"},
                "heartbeat": "60",
                "timeout": "86400",
                "dependencies": "foo",
            },
        ]

    @pytest.fixture
    def task_specs(self, task_spec_dicts):
        """Example deserialised DAG-type workflow tasks specifications."""
        return [_dag.Task.from_spec(s) for s in task_spec_dicts]

    @pytest.fixture
    def spec(self, task_spec_dicts):
        """Example DAG-type workflow specification."""
        return {
            "name": "foo",
            "version": "0.42",
            "description": "A DAGflow",
            "tasks": task_spec_dicts,
            "type": "dag",
        }

    @pytest.fixture
    def instance(self, task_specs):
        """DAG-type workflow specification instance."""
        return seddy_specs.DAGWorkflow("foo", "0.42", task_specs, "A DAGflow")

    def test_init(self, instance, task_specs):
        """Test DAG-type workflow specification initialisation."""
        assert instance.name == "foo"
        assert instance.version == "0.42"
        assert instance.description == "A DAGflow"
        assert instance.task_specs == task_specs
        assert instance.spec_type == "dag"
        assert instance.decisions_builder is seddy_specs.DAGBuilder
        assert instance.dependants == {None: []}

    def test_from_spec(self, spec, task_specs):
        """Test construction from specification."""
        res = seddy_specs.DAGWorkflow.from_spec(spec)
        assert isinstance(res, seddy_specs.DAGWorkflow)
        assert res.name == "foo"
        assert res.version == "0.42"
        assert res.description == "A DAGflow"
        assert res.task_specs == task_specs

    def test_setup(self, instance):
        """Test DAG-type workflow specification pre-computation."""
        instance.setup()
        assert instance.dependants == {
            None: ["foo"],
            "foo": ["bar", "yay"],
            "bar": [],
            "yay": [],
        }



"""Test ``seddy.decider``."""

import os
import socket
from unittest import mock
from concurrent import futures as cf

from seddy import decider as seddy_decider
from seddy import _specs as seddy_specs
import moto
import pytest
from botocore import client as botocore_client


def test_socket_timeout():
    assert socket.getdefaulttimeout() >= 70.0


class TestDecider:
    @pytest.fixture
    def workflow_mocks(self):
        workflows = [
            mock.Mock(spec=seddy_specs.Workflow),
            mock.Mock(spec=seddy_specs.Workflow),
            mock.Mock(spec=seddy_specs.Workflow),
        ]
        workflows[0].name = "spam"
        workflows[0].version = "1.0"
        workflows[1].name = "bar"
        workflows[1].version = "0.42"
        workflows[2].name = "spam"
        workflows[2].version = "1.1"
        return workflows

    @pytest.fixture
    def aws_environment(self):
        env_update = {
            "AWS_DEFAULT_REGION": "us-east-1",
            "AWS_ACCESS_KEY_ID": "id",
            "AWS_SECRET_ACCESS_KEY": "key",
        }
        with mock.patch.dict(os.environ, env_update):
            yield env_update

    @pytest.fixture
    def workflows_spec_file(self, tmp_path):
        return tmp_path / "workflows.json"

    @pytest.fixture
    def instance(self, workflows_spec_file, aws_environment):
        return seddy_decider.Decider(workflows_spec_file, "spam", "eggs", "abcd1234")

    def test_init(self, instance, workflows_spec_file):
        assert instance.workflows_spec_file == workflows_spec_file
        assert instance.domain == "spam"
        assert instance.task_list == "eggs"
        assert isinstance(instance.client, botocore_client.BaseClient)
        assert instance.identity == "abcd1234"

    @moto.mock_swf
    def test_poll_for_decision_task(self, instance):
        # Setup environment
        instance.client.register_domain(
            name="spam", workflowExecutionRetentionPeriodInDays="2"
        )
        instance.client.register_workflow_type(
            domain="spam", name="bar", version="0.42"
        )
        resp = instance.client.start_workflow_execution(
            domain="spam",
            workflowId="1234",
            workflowType={"name": "bar", "version": "0.42"},
            executionStartToCloseTimeout="60",
            taskList={"name": "eggs"},
            taskStartToCloseTimeout="10",
            childPolicy="REQUEST_CANCEL",
        )

        # Run function
        res = instance._poll_for_decision_task()

        # Check result
        assert res == {
            "ResponseMetadata": mock.ANY,
            "events": [
                {
                    "eventId": 1,
                    "eventTimestamp": mock.ANY,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "workflowType": {"name": "bar", "version": "0.42"},
                        "executionStartToCloseTimeout": "60",
                        "taskList": {"name": "eggs"},
                        "taskStartToCloseTimeout": "10",
                        "childPolicy": "REQUEST_CANCEL",
                    },
                },
                {
                    "eventId": 2,
                    "eventTimestamp": mock.ANY,
                    "eventType": "DecisionTaskScheduled",
                    "decisionTaskScheduledEventAttributes": {
                        "startToCloseTimeout": "10",
                        "taskList": {"name": "eggs"},
                    },
                },
                {
                    "eventId": 3,
                    "eventTimestamp": mock.ANY,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {
                        "identity": instance.identity,
                        "scheduledEventId": 2,
                    },
                },
            ],
            "startedEventId": 3,
            "taskToken": mock.ANY,
            "workflowExecution": {"runId": resp["runId"], "workflowId": "1234"},
            "workflowType": {"name": "bar", "version": "0.42"},
        }

    def test_get_workflow(self, instance, workflow_mocks):
        # Setup environment
        load_mock = mock.Mock(return_value=workflow_mocks)
        load_patch = mock.patch.object(seddy_specs, "load_workflows", load_mock)

        # Build input
        task = {
            "ResponseMetadata": mock.ANY,
            "events": [
                {
                    "eventId": 1,
                    "eventTimestamp": mock.ANY,
                    "eventType": "WorkflowExecutionStarted",
                    "workflowExecutionStartedEventAttributes": {
                        "workflowType": {"name": "bar", "version": "0.42"},
                        "executionStartToCloseTimeout": "60",
                        "taskList": {"name": "eggs"},
                        "taskStartToCloseTimeout": "10",
                        "childPolicy": "REQUEST_CANCEL",
                    },
                },
                {
                    "eventId": 2,
                    "eventTimestamp": mock.ANY,
                    "eventType": "DecisionTaskScheduled",
                    "decisionTaskScheduledEventAttributes": {
                        "startToCloseTimeout": "10",
                        "taskList": {"name": "eggs"},
                    },
                },
                {
                    "eventId": 3,
                    "eventTimestamp": mock.ANY,
                    "eventType": "DecisionTaskStarted",
                    "decisionTaskStartedEventAttributes": {
                        "identity": instance.identity,
                        "scheduledEventId": 2,
                    },
                },
            ],
            "previousStartedEventId": 0,
            "startedEventId": 3,
            "taskToken": mock.ANY,
            "workflowExecution": {"runId": mock.ANY, "workflowId": "1234"},
            "workflowType": {"name": "bar", "version": "0.42"},
        }

        # Run function
        with load_patch:
            res = instance._get_workflow(task)

        # Check result
        assert res is workflow_mocks[1]

    def test_get_workflow_unsupported(self, instance, workflow_mocks):
        """Check workflow-get raises for unsupported workflows."""
        # Setup environment
        load_mock = mock.Mock(return_value=workflow_mocks)
        load_patch = mock.patch.object(seddy_specs, "load_workflows", load_mock)

        # Build input
        task = {
            "ResponseMetadata": mock.ANY,
            "taskToken": mock.ANY,
            "workflowExecution": {"runId": mock.ANY, "workflowId": "1234"},
            "workflowType": {"name": "bar", "version": "0.43"},
        }

        # Run function
        with pytest.raises(seddy_decider.UnsupportedWorkflow) as e:
            with load_patch:
                instance._get_workflow(task)

        # Check result
        assert e.value.args[0] == {"name": "bar", "version": "0.43"}

    @moto.mock_swf
    def test_respond_decision_task_completed(self, instance):
        # Setup environment
        instance.client.register_domain(
            name="spam", workflowExecutionRetentionPeriodInDays="2"
        )
        instance.client.register_workflow_type(
            domain="spam", name="bar", version="0.42"
        )
        resp = instance.client.start_workflow_execution(
            domain="spam",
            workflowId="1234",
            workflowType={"name": "bar", "version": "0.42"},
            executionStartToCloseTimeout="60",
            taskList={"name": "eggs"},
            taskStartToCloseTimeout="10",
            childPolicy="REQUEST_CANCEL",
        )
        task = instance.client.poll_for_decision_task(
            domain="spam", identity=instance.identity, taskList={"name": "eggs"}
        )

        # Build input
        decisions = [{"decisionType": "CompleteWorkflowExecution"}]

        # Run function
        instance._respond_decision_task_completed(decisions, task)

        # Check result
        execution_info = instance.client.describe_workflow_execution(
            domain="spam", execution={"workflowId": "1234", "runId": resp["runId"]}
        )
        assert execution_info["executionInfo"]["executionStatus"] == "CLOSED"
        assert execution_info["executionInfo"]["closeStatus"] == "COMPLETED"

    def test_poll_and_run(self, workflow_mocks, aws_environment):
        # Setup environment
        task = {
            "taskToken": "spam",
            "workflowType": {"name": "bar", "version": "0.42"},
            "workflowExecution": {"workflowId": "1234", "runId": "9abc"},
        }

        class Decider(seddy_decider.Decider):
            _poll_for_decision_task = mock.Mock(return_value=task)
            _get_workflow = mock.Mock(return_value=workflow_mocks[1])
            _respond_decision_task_completed = mock.Mock()

        workflow_mocks[1].make_decisions.return_value = [
            {"decisionType": "CompleteWorkflowExecution"}
        ]

        instance = Decider(workflow_mocks, "spam", "eggs")

        # Run function
        instance._poll_and_run()

        # Check calls
        instance._poll_for_decision_task.assert_called_once_with()
        instance._get_workflow.assert_called_once_with(task)
        instance._respond_decision_task_completed.assert_called_once_with(
            [{"decisionType": "CompleteWorkflowExecution"}], task
        )

    def test_poll_and_run_no_result(self, workflow_mocks, aws_environment):
        # Setup environment
        class Decider(seddy_decider.Decider):
            _poll_for_decision_task = mock.Mock(return_value={"taskToken": ""})
            _get_workflow = mock.Mock()
            _respond_decision_task_completed = mock.Mock()

        instance = Decider(workflow_mocks, "spam", "eggs")

        # Run function
        instance._poll_and_run()

        # Check calls
        instance._poll_for_decision_task.assert_called_once_with()
        instance._get_workflow.assert_not_called()
        instance._respond_decision_task_completed.assert_not_called()

    def test_poll_and_run_unsupported(self, workflow_mocks, aws_environment):
        """Task for workflow not in specifications file."""
        # Setup environment
        task = {
            "taskToken": "spam",
            "workflowType": {"name": "bar", "version": "0.43"},
            "workflowExecution": {"workflowId": "1234", "runId": "9abc"},
        }

        class Decider(seddy_decider.Decider):
            _poll_for_decision_task = mock.Mock(return_value=task)
            _get_workflow = mock.Mock(
                side_effect=seddy_decider.UnsupportedWorkflow(task["workflowType"])
            )
            _respond_decision_task_completed = mock.Mock()

        instance = Decider(workflow_mocks, "spam", "eggs")

        # Run function
        with pytest.raises(seddy_decider.UnsupportedWorkflow) as e:
            instance._poll_and_run()
        assert e.value.args[0] == task["workflowType"]

        # Check calls
        instance._poll_for_decision_task.assert_called_once_with()
        instance._get_workflow.assert_called_once_with(task)
        instance._respond_decision_task_completed.assert_not_called()

    def test_poll_and_run_decider_error(self, workflow_mocks, aws_environment):
        """Decision-building raises."""
        # Setup environment
        task = {
            "taskToken": "spam",
            "workflowType": {"name": "bar", "version": "0.42"},
            "workflowExecution": {"workflowId": "1234", "runId": "9abc"},
        }

        class Decider(seddy_decider.Decider):
            _poll_for_decision_task = mock.Mock(return_value=task)
            _get_workflow = mock.Mock(return_value=workflow_mocks[1])
            _respond_decision_task_completed = mock.Mock()

        workflow_mocks[1].make_decisions.side_effect = RuntimeError("malformed specs")

        instance = Decider(workflow_mocks, "spam", "eggs")

        # Build expectation
        exp_decision = {
            "decisionType": "FailWorkflowExecution",
            "failWorkflowExecutionDecisionAttributes": {
                "reason": "RuntimeError",
                "details": "malformed specs",
            },
        }

        # Run function
        with pytest.raises(RuntimeError) as e:
            instance._poll_and_run()
        assert str(e.value) == "malformed specs"

        # Check calls
        instance._poll_for_decision_task.assert_called_once_with()
        instance._get_workflow.assert_called_once_with(task)
        instance._respond_decision_task_completed.assert_called_once_with(
            [exp_decision], task
        )

    def test_run_uncaught(self, workflow_mocks, aws_environment):
        # Setup environment
        class Decider(seddy_decider.Decider):
            _poll_and_run = mock.Mock(side_effect=[None, None, None, KeyboardInterrupt])

        instance = Decider(workflow_mocks, "spam", "eggs")

        # Run function
        with pytest.raises(KeyboardInterrupt):
            instance._run_uncaught()

        # Check calls
        assert instance._poll_and_run.call_args_list == [mock.call()] * 4

    def test_run(self, workflow_mocks, aws_environment):
        # Setup environment
        class Decider(seddy_decider.Decider):
            _run_uncaught = mock.Mock(side_effect=KeyboardInterrupt)

        instance = Decider(workflow_mocks, "spam", "eggs")
        instance._future = mock.Mock(spec=cf.Future)
        instance._future.running.return_value = False

        # Run function
        instance.run()

        # Check calls
        instance._run_uncaught.assert_called_once_with()
        instance._future.result.assert_not_called()

    def test_run_handling_decision(self, workflow_mocks, aws_environment):
        # Setup environment
        class Decider(seddy_decider.Decider):
            _run_uncaught = mock.Mock(side_effect=KeyboardInterrupt)

        instance = Decider(workflow_mocks, "spam", "eggs")
        instance._future = mock.Mock(spec=cf.Future)
        instance._future.running.return_value = True

        # Run function
        instance.run()

        # Check calls
        instance._run_uncaught.assert_called_once_with()
        instance._future.result.assert_called_once_with()


def test_run_app(tmp_path):
    """Ensure decider is run with the correct configuration."""
    # Setup environment
    decider_mock = mock.Mock(spec=seddy_decider.Decider)
    decider_class_mock = mock.Mock(return_value=decider_mock)
    decider_class_patch = mock.patch.object(
        seddy_decider, "Decider", decider_class_mock
    )

    # Build input
    workflows_spec_json = tmp_path / "workflows.json"

    # Run function
    with decider_class_patch:
        seddy_decider.run_app(workflows_spec_json, "spam", "eggs", "abcd1234")

    # Check decider configuration
    decider_class_mock.assert_called_once_with(
        workflows_spec_json, "spam", "eggs", "abcd1234"
    )
    decider_class_mock.return_value.run.assert_called_once_with()



"""SWF decisions making."""

import abc
import enum
import dataclasses
import typing as t


class DeciderError(RuntimeError):
    """Misconfiguration of the decider."""


class ChildPolicy(enum.Enum):
    """Policy for child executions on parent termination.

    .. seealso::
        `StartWorkflowExecution in SWF API documentation
        <https://docs.aws.amazon.com/amazonswf/latest/apireference/API_StartWorkflowExecution.html#SWF-StartWorkflowExecution-request-childPolicy>`_
    """

    TERMINATE = "TERMINATE"
    REQUEST_CANCEL = "REQUEST_CANCEL"
    ABANDON = "ABANDON"


@dataclasses.dataclass
class Registration:
    """Workflow registration configuration.

    Args:
        active: registration status, ``False`` for deprecated
        task_timeout: default decision task time-out (seconds), or "NONE"
            for unlimited
        execution_timeout: default workflow execution time-out (seconds)
        task_list: default decision task-list
        task_priority: default decision task priority
        child_policy: default policy for child executions upon parent
            execution termination
        lambda_role: default IAM role for Lambda access
    """

    active: bool = True
    task_timeout: t.Union[int, str] = None
    execution_timeout: int = None
    task_list: str = None
    task_priority: int = None
    child_policy: ChildPolicy = None
    lambda_role: str = None

    @classmethod
    def from_spec(cls, spec: t.Dict[str, t.Any]):
        """Construct registration configuration from specification.

        Args:
            spec: workflow registration configuration specification
        """

        kw = {}
        if "active" in spec:
            kw["active"] = spec["active"]
        if "task_timeout" in spec:
            kw["task_timeout"] = spec["task_timeout"]
        if "execution_timeout" in spec:
            kw["execution_timeout"] = spec["execution_timeout"]
        if "task_list" in spec:
            kw["task_list"] = spec["task_list"]
        if "task_priority" in spec:
            kw["task_priority"] = spec["task_priority"]
        if "child_policy" in spec:
            kw["child_policy"] = ChildPolicy(spec["child_policy"])
        if "lambda_role" in spec:
            kw["lambda_role"] = spec["lambda_role"]
        return cls(**kw)


class DecisionsBuilder(metaclass=abc.ABCMeta):
    """SWF decision builder.

    Args:
        workflow: workflow specification
        task: decision task
    """

    def __init__(self, workflow: "Workflow", task: t.Dict[str, t.Any]):
        self.workflow = workflow
        self.task = task
        self.decisions = []

    @abc.abstractmethod
    def build_decisions(self):  # pragma: no cover
        """Build decisions from workflow history."""
        raise NotImplementedError


class Workflow(metaclass=abc.ABCMeta):
    """SWF workflow specification.

    Args:
        name: workflow name
        version: workflow version
        registration: workflow registration configuration
    """

    _registration_cls = Registration

    def __init__(
        self,
        name: str,
        version: str,
        description: str = None,
        registration: Registration = None,
    ):
        self.name = name
        self.version = version
        self.description = description
        self.registration = registration

    @classmethod
    def _args_from_spec(
        cls, spec: t.Dict[str, t.Any]
    ) -> t.Tuple[tuple, t.Dict[str, t.Any]]:
        """Construct initialisation arguments from workflow specification.

        Args:
            spec: workflow specification

        Returns:
            initialisation positional and keyword arguments
        """

        args = (spec["name"], spec["version"])
        kwargs = {}
        if "description" in spec:
            kwargs["description"] = spec["description"]
        if "registration" in spec:
            kwargs["registration"] = cls._registration_cls.from_spec(
                spec["registration"]
            )
        return args, kwargs

    @classmethod
    def from_spec(cls, spec: t.Dict[str, t.Any]):
        """Construct workflow type from specification.

        Args:
            spec: workflow specification
        """

        args, kwargs = cls._args_from_spec(spec)
        return cls(*args, **kwargs)

    @property
    @abc.abstractmethod
    def decisions_builder(self) -> DecisionsBuilder:  # pragma: no cover
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def spec_type(self) -> str:  # pragma: no cover
        raise NotImplementedError

    def setup(self):
        """Set up workflow specification.

        Useful for pre-calculation or other initialisation.
        """

    def make_decisions(self, task: t.Dict[str, t.Any]) -> t.List[t.Dict[str, t.Any]]:
        """Build decisions from workflow history.

        Args:
            task: decision task

        Returns:
            workflow decisions
        """

        builder = self.decisions_builder(self, task)
        builder.build_decisions()
        return builder.decisions


def make_decisions_on_error(exception: Exception) -> t.List[t.Dict[str, t.Any]]:
    """Build workflow-fail decision on decider exception.

    Args:
        exception: decider exception being handled

    Returns:
        workflow-fail decision
    """

    decision_attrs = {"reason": exception.__class__.__name__}
    message = str(exception)
    if message:
        decision_attrs["details"] = message
    decision = {
        "decisionType": "FailWorkflowExecution",
        "failWorkflowExecutionDecisionAttributes": decision_attrs,
    }
    return [decision]



"""Workflows specs serialisation and desieralisation."""

import json
import pathlib
import typing as t
import logging as lg

from . import Workflow

logger = lg.getLogger(__package__)


def _construct_workflows(workflows_spec: t.Dict[str, t.Any]) -> t.List[Workflow]:
    """Construct workflows from specification.

    Args:
        workflows_spec: workflows specifications

    Returns:
        workflow type specifications
    """

    from . import WORKFLOW

    assert (1,) < tuple(map(int, workflows_spec["version"].split("."))) < (2,)
    workflows = []
    for workflow_spec in workflows_spec["workflows"]:
        workflow_cls = WORKFLOW[workflow_spec["spec_type"]]
        workflow = workflow_cls.from_spec(workflow_spec)
        workflows.append(workflow)
    return workflows


def _load_specs(workflows_file: pathlib.Path) -> t.Dict[str, t.Any]:
    """Load workflows specifications file.

    Determines load method from the file suffix. Supported file types:

    * JSON
    * YAML

    Args:
        workflows_file: workflows specifications file path

    Returns:
        workflows specifications
    """

    logger.info("Loading workflows specifictions from '%s'", workflows_file)
    workflows_text = workflows_file.read_text()
    if workflows_file.suffix == ".json":
        return json.loads(workflows_text)
    elif workflows_file.suffix in (".yml", ".yaml"):
        try:
            import yaml
        except ImportError as e:
            try:
                from ruamel import yaml
            except ImportError as er:
                raise e from er
        return yaml.safe_load(workflows_text)
    raise ValueError("Unknown extension: %s" % workflows_file.suffix)


def load_workflows(workflows_file: pathlib.Path) -> t.List[Workflow]:
    """Load workflows specifications file.

    Determines load method from the file suffix. Supported file types:

    * JSON
    * YAML

    Args:
        workflows_file: workflows specifications file path

    Returns:
        workflows specifications
    """

    workflows_specs = _load_specs(workflows_file)
    return _construct_workflows(workflows_specs)
